# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
""" ``PDBDataFrame`` as a subclass of ``Pandas DataFrame``.

Several features are added to make PDB data more accessible and selectable:

1. Properties like ``sequences``, ``heavy_atoms``, ``backbone``, and ``water`` are
directly accessed by ``.`` operation.

2. Atom selection by using methods whose names are just the column names plus ``s``
(plural form). For example, selecting atoms by names is simply
``df.atom_names([names])`` where ``atom_name`` is the column name
and ``atom_names`` is the selection function. Each selection returns a
``PDBDataFrame`` object as well, which means we can chain selections one by one
like ``df.atom_names([names]).residue_numbers([numbers])``.

3. Distance matrix as a ``@property`` and ``@classmethod``.
"""
from __future__ import annotations

import functools
import warnings
from collections import defaultdict
from collections.abc import Iterable

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from scipy.spatial.distance import cdist, pdist, squareform  # type: ignore
from scipy.spatial.transform import Rotation  # type: ignore
from typing_extensions import Self

from .constants import AMINO_ACIDS, ELEMENT_MASSES

RESIDUE_CODES = AMINO_ACIDS
"""dict[str, str], turn 3-, 2-, and 1-letter residue codes to 1-letter codes."""


PDBX_COLS = {
    "record_name": "group_PDB",
    "atom_number": "id",
    "atom_name": "label_atom_id",
    "alt_loc": "label_alt_id",
    "residue_name": "label_comp_id",
    "chain_id": "label_asym_id",
    "residue_number": "label_seq_id",
    "insertion": "pdbx_PDB_ins_code",
    "x_coord": "Cartn_x",
    "y_coord": "Cartn_y",
    "z_coord": "Cartn_z",
    "occupancy": "occupancy",
    "b_factor": "B_iso_or_equiv",
    "segment_id": "label_entity_id",
    "element_symbol": "type_symbol",
    "charge": "pdbx_formal_charge",
    "nmr_model": "pdbx_PDB_model_num",
}
"""dict[str, str], PDB and mmCIF column name dictionary."""


class PDBDataFrame(pd.DataFrame):
    """Pandas DataFrame with extended attributes and methods for PDB data.

    It enables Pythonic atom selection methods and convenient ``.`` accessing to common
    PDB structure properties.

    Args:
        *args: all ``pd.DataFrame`` positional arguments. For example, the
            ``_atom_site`` dataframe returned by reading a PDB file.
        pdb_format (optional): PDB format in the underlying provided data.
            If ``None``, ``PDB`` is assumed. Defaults to **None**.
        use_squared_distance (optional): whether to use squared distance
            when calculating distance matrix. Defaults to **True**.
        use_square_form (optional): whether to use a square matrix
            for the distance matrix. Defaults to **False**.
        **kwargs: all ``pd.DataFrame`` acceptable keyword arguments.

    Returns:
        A ``PDBDataFrame`` instance.

    Example
    -------
    >>> from pdbx2df import read_pdb, PDBDataFrame
    >>> pdb = read_pdb(pdb_id='1vii')
    >>> pdb_df = pdb['_atom_site']
    >>> pdb_df = PDBDataFrame(pdb_df)

    Warnings
    --------
    This subclass uses a custom ``__hash__`` function for caching some calculations.
    And thus a custom ``__eq__`` function is also implemented. For other typical
    ``DataFrame`` operations, use those ``.all()``, ``.any()``, ``.bool()`` functions
    to do comparison.

    """

    _metadata = [
        "_use_squared_distance",
        "_use_square_form",
        "_is_chimera",
        "_RESIDUE_CODES",
        "_ELEMENT_MASSES",
        "_pdb_format",
    ]

    def __init__(
        self,
        *args,
        pdb_format: str | None = None,
        use_squared_distance: bool = True,
        use_square_form: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._pdb_format: str | None = pdb_format
        if self._pdb_format is None:
            self._pdb_format = "PDB"
        if self._pdb_format.lower() in ["mmcif", "pdbx"]:
            self._pdbx_to_pdb()
        self._use_squared_distance: bool = use_squared_distance
        self._use_square_form: bool = use_square_form
        self._hash_random_state: int = 0
        self._is_chimera = False
        self._RESIDUE_CODES: dict[str, str] = {}
        self._ELEMENT_MASSES: dict[str, float] = {}
        self._ter_line_removed: bool = False
        self._atoms: Self | None = None

    @property
    def _constructor(self):
        return PDBDataFrame

    def __hash__(self) -> int:
        """Uses head X coords to hash; for distance matrix calculation cache."""
        try:
            sample_atom_numbers = self.sample(
                5, random_state=self.hash_random_state, replace=True
            )["atom_number"]
        except ValueError:
            sample_atom_numbers = []
        return hash(tuple(self[self.atom_number.isin(sample_atom_numbers)].x_coord))

    def __eq__(self, other) -> bool:
        """Uses head X coords to compare; for distance matrix calculation cache."""
        return self.__hash__() == other.__hash__()

    def _pdbx_to_pdb(self, keep_original: bool = False):
        """Converts PDBx '_atom_site' DataFrame to PDB format.

        Args:
            keep_original (optional): whether to keep the original columns in the PDBx
                '_atom_site' DataFrame. Defaults to **False**.
        """
        for pdb_name, pdbx_name in PDBX_COLS.items():
            self[pdb_name] = self[pdbx_name]
        if not keep_original:
            drop_columns = [col for col in self.columns if col not in PDBX_COLS.keys()]
            self.drop(columns=drop_columns, inplace=True)
        self._pdb_format = "PDBx"
        self = self[list(PDBX_COLS.keys())]

    @property
    def pdb_format(self) -> str:
        """
        The format of the current PDBDataFrame.
        """
        return self._pdb_format  # type: ignore

    @property
    def RESIDUE_CODES(self) -> dict[str, str]:
        """
        A dict of ``residue_name`` as keys and ``residue_code`` as values, where
        ``residue_code`` is a 1-character code used in sequences. **Settable**.
        """
        if not self._RESIDUE_CODES:
            res_name_width = 3
            if self.is_chimera:
                res_name_width = 4
            self._RESIDUE_CODES = {
                res.upper().ljust(res_name_width): code
                for res, code in RESIDUE_CODES.items()
            }

        return self._RESIDUE_CODES

    @RESIDUE_CODES.setter
    def RESIDUE_CODES(self, residue_codes: dict[str, str]) -> None:
        res_name_width = 3
        if self.is_chimera:
            res_name_width = 4

        self._RESIDUE_CODES = {
            res.upper().ljust(res_name_width): code
            for res, code in residue_codes.items()
        }

    @property
    def ELEMENT_MASSES(self) -> dict[str, float]:
        """
        A dict of ``element_symbol`` as keys and ``element_mass`` as values, where
        ``element_mass`` is taken from NIST. **Settable**.
        """
        if not self._ELEMENT_MASSES:
            self._ELEMENT_MASSES = {
                e.upper().rjust(2): mass for e, mass in ELEMENT_MASSES.items()
            }
        return self._ELEMENT_MASSES

    @ELEMENT_MASSES.setter
    def ELEMENT_MASSES(self, element_masses: dict[str, float]):
        self._ELEMENT_MASSES = {
            e.upper().rjust(2): mass for e, mass in element_masses.items()
        }

    @property
    def is_chimera(self) -> bool:
        """
        Whether the original read-in PDB was Chimera compatible format.
            The main effect is the ``residue_name`` str width is 4 in the
            Chimera compatible format instead of 3 as in the standard PDB
            format.  **Not settable**.
        """
        try:
            if len(self.head(1).residue_name[0]):
                self._is_chimera = True
        except AttributeError:
            pass  # 'residue_name' not in self.coords
        except IndexError:
            pass  # empty

        return self._is_chimera

    @property
    def hash_random_state(self) -> int:
        """The ``random_state`` used in the ``__hash__`` function. **Settable**."""
        return self._hash_random_state

    @hash_random_state.setter
    def hash_random_state(self, random_state: int) -> None:
        self._hash_random_state = random_state

    @property
    def use_squared_distance(self) -> bool:
        """
        Whether R or R^2 is used in distance matrix calculations.
            Using R^2 saves computation time. **Settable**.
        """
        return self._use_squared_distance

    @use_squared_distance.setter
    def use_squared_distance(self, use_r2: bool) -> None:
        self._use_squared_distance = use_r2

    @property
    def use_square_form(self) -> bool:
        """
        Whether the distance matrix will be in a square form.
            Using square form consumes less memory. **Settable**.
        """
        return self._use_square_form

    @use_square_form.setter
    def use_square_form(self, square_form: bool) -> None:
        self._use_square_form = square_form

    @property
    def atoms(self) -> Self:
        """Gets atoms in the ``ATOM`` and ``HETATM`` entries.
        In other words, removing 'TER' lines.

        Returns:
            sub ``PDBDataFrame``.
        """
        if self._atoms is None:
            if self._ter_line_removed:
                self._atoms = self
            else:
                self._atoms = self.record_names(["ATOM  ", "HETATM"])
                self._ter_line_removed = True
        return self._atoms

    @property
    def coords(self) -> Self:
        """
        Gets the ``x_coord``, ``y_coord``, and ``z_coord`` columns only.
            Use ``pdb_df.coords.values`` to get the underlying
            Numpy array of the coordinates.
        """
        return self.atoms[["x_coord", "y_coord", "z_coord"]]

    @property
    @functools.lru_cache()
    def sequences(self) -> dict[str, str]:
        """
        Gets the sequences for each chain as a dict of ``chain_id``
            as key(s) and ``chain_sequence`` as value(s).
        """
        chain_sequence: dict[str, str] = defaultdict(str)
        for resi_info in self.residue_list:
            chain_sequence[resi_info[0]] += self.RESIDUE_CODES[resi_info[1]]
        return chain_sequence

    @property
    def residue_list(self) -> list[tuple]:
        """
        Gets all residues as a list of tuple
            (``chain_id``, ``residue_name``, ``residue_number``)
        """
        return PDBDataFrame.get_residue_list(self)

    @property
    def backbone(self) -> Self:
        """Gets backbone or N+CA+C+O atoms.

        Returns:
            sub ``PDBDataFrame``
        """
        return self.residues.atom_names(
            ["N", "CA", "C", "O"],
            suppress_warning=True,
        ).element_symbols(["C", "N", "O"])

    @property
    def side_chain(self) -> Self:
        """Gets side chain or NOT N+CA+C+O atoms.

        Returns:
            sub ``PDBDataFrame``.
        """
        return self.residues.atom_names(
            ["N", "CA", "C", "O"],
            invert=True,
            suppress_warning=True,
        ).element_symbols(["C", "N", "O", "S"])

    @property
    def ca_atoms(self) -> Self:
        """Gets the alpha carbon (CA) atoms.

        Returns:
            sub ``PDBDataFrame``.
        """
        return self.residues.atom_names(
            ["CA"],
            suppress_warning=True,
        ).element_symbols(["C"])

    @property
    def heavy_atoms(self) -> Self:
        """Gets the heavy or NOT hydrogen atoms.

        Returns:
            sub ``PDBDataFrame``.
        """
        return self.element_symbols(["H", "D", "T"], invert=True)

    @property
    def hetero_atoms(self) -> Self:
        """Gets the hetero (``HETATM``) atoms.

        Returns:
            sub ``PDBDataFrame``.
        """
        return self.record_names(["HETATM"])

    @property
    def residues(self) -> Self:
        """Gets the residue (``ATOM``) atoms.

        Returns:
            sub ``PDBDataFrame``.
        """
        return self.record_names(["ATOM  "])

    @property
    def water(self) -> Self:
        """Gets all water atoms.

        Returns:
            sub ``PDBDataFrame``.
        """
        return self.atoms.residue_names(["HOH"])

    @property
    def n_atoms(self) -> int:
        """Gets the number of atoms."""
        return len(self.atoms)

    @property
    def n_residues(self) -> int:
        """Gets the number of residues."""
        return len(self.residues)

    @property
    def n_chains(self) -> int:
        """Gets the number of chains."""
        ter_lines = self[self.record_name == "TER   "]
        ter_residues = PDBDataFrame.get_residue_list(ter_lines)

        oxt_lines = self[self.atom_name == " OXT"]
        oxt_residues = PDBDataFrame.get_residue_list(oxt_lines)

        n_chain_ids = len(self.chain_id.unique())

        ter_oxt_residues = ter_residues
        for oxt_residue in oxt_residues:
            if oxt_residue not in ter_oxt_residues:
                ter_oxt_residues.append(oxt_residue)

        return max(len(ter_oxt_residues), n_chain_ids)

    @property
    def n_segments(self) -> int:
        """Gets the number of segments."""
        return len(self.atoms.segment_id.unique())

    @property
    def n_models(self) -> int:
        """Gets the number of models."""
        if "nmr_model" in self.columns:
            return len(self.nmr_model.unique())
        return 1

    @property
    @functools.lru_cache()
    def center_of_geometry(self) -> np.ndarray:
        """Gets the center of geometry as a ``(3, )`` ``np.ndarray``."""
        return np.mean(self.coords.values, axis=0)

    @property
    @functools.lru_cache()
    def center_of_mass(self) -> np.ndarray:
        """Gets the center of mass as a ``(3, )`` ``np.ndarray``."""
        masses = self.atoms.get_masses()
        masses = masses / masses.sum()
        return np.sum(self.coords.values * masses[:, None], axis=0)

    @property
    @functools.lru_cache()
    def radius_of_gyration(self) -> float:
        """Gets the radius of gyration"""
        com = self.center_of_mass
        com_t = (com[0], com[1], com[2])  # type: ignore
        dist_to_com = PDBDataFrame.get_distance_matrix(self.atoms, com_t, use_r2=True)
        masses = self.atoms.get_masses()
        masses = masses / masses.sum()
        return np.sum(dist_to_com * masses[:, None], axis=0)[0]

    @functools.lru_cache()
    def get_masses(self) -> np.ndarray:
        """Gets the masses for all atoms in the current dataframe."""
        masses = np.zeros(len(self.atoms), dtype="float32")
        for i, element in enumerate(self.atoms.element_symbol):
            masses[i] = self.ELEMENT_MASSES[element]
        return masses

    @property
    def distance_matrix(self) -> np.ndarray:
        """Gets the distance matrix."""
        return PDBDataFrame.get_distance_matrix(
            self.atoms,
            use_r2=self.use_squared_distance,
            square_form=self.use_square_form,
        )

    def rmsd(
        self,
        other: Self | np.ndarray | None = None,
        align: bool = True,
        weights: list | None = None,
        selection: Self | list | None = None,
    ) -> list | float:
        """Calculates RMSD 1) among sets of coordinates in one ``PDBDataFrame`` with
        multiple ``nmr_model`` s or 2) two sets of coordinates in two ``PDBDataFrames``.

        Args:
            other (optional): the other ``PDBDataFrame`` or (N, 3) ``numpy.ndarray``
                to calculate RMSD against. If ``None``, ``self`` should contain at least
                two sets of coordinates (``nmr_model`` has >= 2 unique values).
                Defaults to **None**.
            align (optional): whether to align the structures before calculating RMSD.
                If ``False``, the ``weights`` and ``selection`` keywords are ignored.
                Defaults to **True**.
            weights (optional): a list of weights for all the atoms in ``selection`` to
                do structure alignment. If ``None``, all coordinates in the
                ``selection`` or ``self`` have the same weights. Defaults to **None**.
            selection (optional): a list of ``atom_number`` s in ``self`` or a
                PDBDataFrame after the filtering methods.
                If ``None``, all coordinates in ``self`` are
                used for structure alignment. Defaults to **None**.

        Returns:
            RMSD or a list of RMSD's.

        Raises:
            ValueError: if dimensionalities mismatch among ``self``, ``other``,
                ``weights``, and ``selection`` if they are not ``None``;
                or ``atom_number`` s in ``self`` are not unique.
            TypeError: if ``other``, ``weights``, and ``selection`` have unsupported
                types.
        """
        result: list = []

        other_coords_list: list = []
        if "nmr_model" in self.columns and len(self.nmr_model.unique()) >= 2:
            all_models = self.nmr_model.unique()
            first_model = self.nmr_models(int(all_models[0]))
            if len(first_model.atom_number.unique()) != len(first_model):
                raise ValueError("'atom_number's in 'self' are not unique.")
            n_atoms = len(first_model)
            if other is None:
                for other_index in all_models[1:]:
                    other_model = self.nmr_models(int(other_index))
                    other_coords_list.append(other_model.coords.values)
        else:
            first_model = self
            if len(first_model.atom_number.unique()) != len(first_model):
                raise ValueError("'atom_number's in 'self' are not unique.")
            n_atoms = len(first_model)
            if other is None:
                message = "'self' has only one set of coordinates and 'other' "
                message += "is not provided."
                raise ValueError(message)
                # message = "'other' is not a PDBDataFrame or np.ndarray instance "
                # message += f"but a {type(other)}."
                # raise ValueError(message)
            elif isinstance(other, type(self)):
                if (self.atom_number != other.atom_number).any():
                    message = "'self' and 'other' have mismatched 'atom_number' values."
                    raise ValueError(message)
                other_coords_list.append(other.coords.values)
            elif isinstance(other, np.ndarray):
                if other.shape != (n_atoms, 3):
                    message = f"'other' shape is {other.shape} but expected to be"
                    message += f" ({n_atoms}, 3)."
                    raise ValueError(message)
                other_coords_list.append(other)
            else:
                raise TypeError(f"Unsupported type {type(other)} for 'other'.")

        first_coords = first_model.coords.values
        align_weights = np.ones(n_atoms)
        if align:
            if selection is not None:
                align_weights = np.zeros(n_atoms)
                if not isinstance(selection, (type(self), list)):
                    message = f"'selection' type is {type(selection)}, but list"
                    message += " or PDBDataFrame is expected."
                    raise TypeError(message)
                n_selection = len(selection)
                if len(selection) > n_atoms:
                    message = f"'selection' length is {len(selection)}, "
                    message += f"larger than the number of atoms {n_atoms}."
                    raise ValueError(message)
                if isinstance(selection, type(self)):
                    selection = self.atom_number.to_list()
                mask = first_model.atom_number.isin(selection)
                align_weights[mask] = 1.0

                if weights is not None:
                    if not isinstance(weights, list):
                        message = f"'weights' type is {type(weights)}, but list"
                        message += " is expected."
                        raise TypeError(message)
                    if len(weights) != n_selection:
                        message = f"'weights' length is {len(weights)}, "
                        message += "not equal to the selection length"
                        message += f" {n_selection}."
                        raise ValueError(message)
                    align_weights[mask] = weights
            elif weights is not None:
                if not isinstance(weights, list):
                    message = f"'weights' type is {type(weights)}, but list"
                    message += " is expected."
                    raise TypeError(message)
                if len(weights) != n_atoms:
                    message = f"'weights' length is {len(weights)}, "
                    message += "not equal to the number of atoms"
                    message += f" {n_atoms} without 'selection'."
                    raise ValueError(message)
                align_weights = weights

        first_centered = False
        for other_coords in other_coords_list:
            if align:
                if not first_centered:
                    first_coords = first_coords - first_model.center_of_geometry
                    first_centered = True
                other_cog = np.mean(other_coords, axis=0)
                other_coords = other_coords - other_cog
                rot = Rotation.align_vectors(
                    first_coords, other_coords, weights=align_weights
                )[0]
                other_coords = rot.apply(other_coords)

            rms = np.sqrt(np.sum((first_coords - other_coords) ** 2) / n_atoms)
            result.append(rms)

        if len(result) == 1:
            return result[0]

        return result

    def record_names(self, names: list[str], invert: bool = False) -> Self:
        """Filter by ``record_name``.

        Args:
            names (required): a list of ``record_name`` s.
            invert (optional): whether to invert the selection. Defaults to **False**.
        Returns:
            sub ``PDBDataFrame``
        """
        names = [name.strip().upper() for name in names]
        if self.pdb_format.upper() == "PDB":  # type: ignore
            names = [name.ljust(6) for name in names]
        if invert:
            return self[~self.record_name.isin(names)]
        return self[self.record_name.isin(names)]

    def atom_numbers(
        self,
        numbers: list[int] | int,
        relation: str | None = None,
        invert: bool = False,
    ) -> Self:
        """Filter by ``atom_number``.

        Args:
            numbers (required): one or a list of ``atom_number`` s.
            relation (optional): ``atom_number`` relationship to ``numbers``.
                If ``numbers`` is an integer, it has to be one of ``<``, ``<=``, ``=``,
                ``>=``, and ``>``. If ``None``, ``<=`` is used. Ignored if a list of
                integers are provided to ``numbers``. Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        return self._filter_num_col(
            numbers, "atom_number", relation=relation, invert=invert
        )

    def atom_names(
        self,
        names: list[str],
        names_2c: list[str] | None = None,
        invert: bool = False,
        suppress_warning: bool = False,
    ) -> Self:
        """Filter by ``atom_name``.

        Args:
            names (required): a list of ``atom_name`` s whose ``element_symbols`` have
                only one character. Atoms in common residues and ligands should be
                provide here like ``C, H, O, N, S, P, F``.
            names_2c (optional): a list of ``atom_name`` s whose ``element_symbols``
                have two characters like ion (``FE``) and chloride (``CL``).
                Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.
            suppress_warning: whether to suppress the warning message about
                possible conflicts between ``names`` and ``names_2c``.
                Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        atom_name_strings = [name.strip().upper() for name in names]
        if self.pdb_format.upper() == "PDB":  # type: ignore
            atom_name_strings = []
            for name in names:
                if len(name) == 4:
                    atom_name_strings.append(name)
                else:
                    atom_name_strings.append(f" {name}".ljust(4))
                if (
                    len(name) == 2
                    and f"{name[0]}{name[1].lower()}" in self.ELEMENT_MASSES
                ):
                    if f" {name[0]}" not in self.ELEMENT_MASSES:
                        atom_name_strings.append(f"{name}".ljust(4))
                        # eg 'MG' where ' M' is not a legal element in self.ELEMENT_MASSES.
                        continue
                    if suppress_warning:
                        continue
                    message = f"Atom name {name} is an atom of element {name[0]} "
                    message += f"but not element {name[0]}{name[1].lower()}."
                    message += "If you want the latter, put it in the 'names_2c' list."
                    warnings.warn(
                        message,
                        RuntimeWarning,
                        stacklevel=2,
                    )
            if names_2c is not None:
                for name in names_2c:
                    atom_name_strings.append(f"{name}".ljust(4))

        if invert:
            return self[~self.atom_name.isin(atom_name_strings)]
        return self[self.atom_name.isin(atom_name_strings)]

    def alt_locs(self, locs: list[str], invert: bool = False) -> Self:
        """Filter by ``alt_loc``.

        Args:
            locs (required): a list of ``alt_loc`` s.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub PDBDataFrame
        """
        if invert:
            return self[~self.alt_loc.isin(locs)]
        return self[self.alt_loc.isin(list(locs) + [" "])]

    def residue_names(self, names: list[str], invert: bool = False) -> Self:
        """Filter by ``residue_names``.

        Args:
            names (required): a list of ``residue_name`` s
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        names = [name.strip().upper() for name in names]
        if self.pdb_format.upper() == "PDB" and self.is_chimera:  # type: ignore
            names = [(name + " ").upper().rjust(4) for name in names]
        if invert:
            return self[~self.residue_name.isin(names)]
        return self[self.residue_name.isin(names)]

    def chain_ids(self, ids: list[str], invert: bool = False) -> Self:
        """Filter by ``chain_id``.

        Args:
            ids (required): a list of ``chain_id`` s.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        if invert:
            return self[~self.chain_id.isin(ids)]
        return self[self.chain_id.isin(ids)]

    def residue_numbers(
        self,
        numbers: list[int] | int,
        relation: str | None = None,
        invert: bool = False,
    ) -> Self:
        """Filter by ``residue_number``.

        Args:
            numbers (required): one or a list of ``residue_number`` s.
            relation (optional): ``residue_number`` relationship to ``numbers``.
                If ``numbers`` is an integer, it has to be one of ``<``, ``<=``, ``=``,
                ``>=``, and ``>``. If ``None``, '<=' is used. Ignored if a list of
                integers are provided to ``numbers``. Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        return self._filter_num_col(
            numbers, "residue_number", relation=relation, invert=invert
        )

    def insertions(self, codes: list[str], invert: bool = False) -> Self:
        """Filter by ``insertion``.

        Args:
            codes (required): a list of ``insertion`` codes.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        if invert:
            return self[~self.insertion.isin(codes)]
        return self[self.insertion.isin(codes)]

    def x_coords(
        self,
        value: float,
        relation: str | None = None,
        invert: bool = False,
        epsilon: float = 0.01,
    ) -> Self:
        """Filter by ``x_coord``.

        Args:
            value (required): value to select ``x_coord`` s.
            relation (optional): ``x_coord`` relationship to ``value``.
                It has to be one of ``'<'``, ``'<='``, ``'='``, ``'>='``, and ``'>'``.
                If ``None``, ``'<='`` is used. Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.
            epsilon (optional): atoms ``abs(x_coord - value)`` <= ``epsilon``
                are selected when ``invert`` = ``False``. Defaults to **0.01**.

        Returns:
            sub ``PDBDataFrame``
        """
        return self._filter_num_col(
            value, "x_coord", relation=relation, invert=invert, epsilon=epsilon
        )

    def y_coords(
        self,
        value: float,
        relation: str | None = None,
        invert: bool = False,
        epsilon: float = 0.01,
    ) -> Self:
        """Filter by ``y_coord``.

        Args:
            value (required): value to select ``y_coord`` s.
            relation (optional): ``y_coord`` relationship to ``value``.
                It has to be one of ``'<'``, ``'<='``, ``'='``, ``'>='``, and ``'>'``.
                If ``None``, ``'<='`` is used. Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.
            epsilon (optional): atoms ``abs(y_coord - value)`` <= ``epsilon``
                are selected when ``invert`` = ``False``. Defaults to **0.01**.

        Returns:
            sub ``PDBDataFrame``
        """
        return self._filter_num_col(
            value, "y_coord", relation=relation, invert=invert, epsilon=epsilon
        )

    def z_coords(
        self,
        value: float,
        relation: str | None = None,
        invert: bool = False,
        epsilon: float = 0.01,
    ) -> Self:
        """Filter by ``z_coord``.

        Args:
            value (required): value to select ``z_coord`` s.

            relation (optional): ``z_coord`` relationship to ``value``.
                It has to be one of ``'<'``, ``'<='``, ``'='``, ``'>='``, and ``'>'``.
                If ``None``, ``'<='`` is used. Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.
            epsilon (optional): atoms ``abs(z_coord - value)`` <= ``epsilon``
                are selected when ``invert`` = ``False``. Defaults to **0.01**.

        Returns:
            sub ``PDBDataFrame``
        """
        return self._filter_num_col(
            value, "z_coord", relation=relation, invert=invert, epsilon=epsilon
        )

    def occupancies(
        self,
        value: float,
        relation: str | None = None,
        invert: bool = False,
        epsilon: float = 0.01,
    ) -> Self:
        """Filter by ``occupancy``.

        Args:
            value (required): value to select ``occupancy`` s.
            relation (optional): ``occupancy`` relationship to ``value``.
                It has to be one of ``'<'``, ``'<='``, ``'='``, ``'>='``, and ``'>'``.
                If ``None``, ``'<='`` is used. Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.
            epsilon (optional): atoms ``abs(occupancy - value)`` <= ``epsilon``
                are selected when ``invert`` = ``False``. Defaults to **0.01**.

        Returns:
            sub ``PDBDataFrame``
        """
        return self._filter_num_col(
            value, "occupancy", relation=relation, invert=invert, epsilon=epsilon
        )

    def b_factors(
        self,
        value: float,
        relation: str | None = None,
        invert: bool = False,
        epsilon: float = 0.01,
    ) -> Self:
        """Filter by ``b_factor``.

        Args:
            value (required): value to select ``b_factor`` s.
            relation (optional): ``b_factor`` relationship to ``value``.
                It has to be one of ``'<'``, ``'<='``, ``'='``, ``'>='``, and ``'>'``.
                If ``None``, ``'<='`` is used. Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.
            epsilon (optional): atoms ``abs(b_factor - value)`` <= ``epsilon``
                are selected when ``invert`` = ``False``. Defaults to **0.01**.

        Returns:
            sub ``PDBDataFrame``
        """
        return self._filter_num_col(
            value, "b_factor", relation=relation, invert=invert, epsilon=epsilon
        )

    def segment_ids(self, ids: list[str], invert: bool = False) -> Self:
        """Filter by ``segment_id``.

        Args:
            ids (required): a list of ``segment_id`` s.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        ids = [i.strip() for i in ids]
        if self.pdb_format.upper() == "PDB":  # type: ignore
            ids = [i.ljust(4) for i in ids]
        if invert:
            return self[~self.segment_id.isin(ids)]
        return self[self.segment_id.isin(ids)]

    def element_symbols(self, symbols: list[str], invert: bool = False) -> Self:
        """Filter by ``element_symbol``.

        Args:
            symbols (required): a list of ``element_symbol`` s.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        symbols = [symbol.strip().upper() for symbol in symbols]
        if self.pdb_format.upper() == "PDB":  # type: ignore
            symbols = [symbol.rjust(2) for symbol in symbols]
        if invert:
            return self[~self.element_symbol.isin(symbols)]
        return self[self.element_symbol.isin(symbols)]

    def charges(self, charges: list[str], invert: bool = False) -> Self:
        """Filter by ``charge``.

        Args:
            charges (required): a list of ``charge`` s.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``

        Notes: ``charge`` is ``2-char`` string in the PDB specifications.
        """
        charges = [charge.strip() for charge in charges]
        if self.pdb_format.upper() == "PDB":  # type: ignore
            charges = [charge.rjust(2) for charge in charges]
        if invert:
            return self[~self.charge.isin(charges)]
        return self[self.charge.isin(charges)]

    def nmr_models(
        self, models: list[int] | int, relation: str | None = None, invert: bool = False
    ) -> Self:
        """Filter by ``nmr_model``.

        Args:
            models (required): one or a list of ``nmr_model`` ids.
            relation (optional): ``nmr_model`` relationship to ``models``.
                If ``models`` is an integer, it has to be one of ``'<'``, ``'<='``,
                ``'='``, ``'>='``, and ``'>'``. If ``None``, ``'<='`` is used.
                Ignored if a list of integers are provided to ``models``.
                Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        if "nmr_model" in self.columns:
            return self._filter_num_col(
                models, "nmr_model", relation=relation, invert=invert
            )
        return self

    def distances(
        self,
        other: np.ndarray | Self | Iterable,
        cut_off: float = np.inf,
        to: str | None = None,
        invert: bool = False,
    ) -> Self:
        """Filter by ``distance`` to a reference point or group of atoms.

        Args:
            other: the other group's coordinate(s).
            cut_off: the distance cutoff to filter.
            to: if ``other`` is a group atoms, using which method to determine
                whether the atoms meet the cut_off distance. If None, ``COM`` or
                center of mass is used if ``other`` is ``PDBDataFrame``, and ``COG``
                or center of geometry is used if ``other`` is ``np.ndarray`` or
                ``Iterable``. The following are allowed:

                    ``com``, ``center of mass``, ``center_of_mass``:
                        use the center of mass for the ``other``.
                    ``cog``, ``center of geometry``, ``center_of_geometry``:
                        use the center of geometry for the ``other``.
                    ``all``:
                        whether all the pair-distances meet the ``cut_off``
                        distance criteria.
                    ``any``:
                        whether any of the pair-distances meets the ``cut_off``
                        distance criteria.
            invert: whether to invert the selection. Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``
        """
        if not (isinstance(other, type(self)) or isinstance(other, Iterable)):
            message = "Only 'PDBDataFrame', 'np.ndarray', and 'Iterable' types "
            message += f"are supported, not {type(other)} for 'other'."
            raise TypeError(message)

        if isinstance(other, Iterable) and not isinstance(other, type(self)):
            other = np.asanyarray(other, dtype="float32")
            if other.shape != (3,):
                if len(other.shape) != 2 or other.shape[1] != 3:
                    message = "An 'Iterable' input should have a (N, 3) or (3,) shape, "
                    message += f"not '{other.shape}' in 'other'."
                    raise TypeError(message)
            else:
                other = other.reshape(1, 3)

        cut_off = cut_off**2

        allowed_tos = [
            "com",
            "center of mass",
            "center_of_mass",
            "cog",
            "center of geometry",
            "center_of_geometry",
            "any",
            "all",
        ]
        if to is None or to.lower() not in allowed_tos:
            if to is not None:
                message = "Only center of mass (COM), center of geometry (COG) "
                message += f"'all', or 'any' is supported. '{to}' is reset to "
                if isinstance(other, type(self)):
                    message += "'COM' for 'to'."
                else:
                    message += "'COG' for 'to'."
                warnings.warn(
                    message,
                    RuntimeWarning,
                    stacklevel=2,
                )
            to = "COG"
            if isinstance(other, type(self)):
                to = "COM"
        elif to.lower() in [
            "com",
            "center of mass",
            "center_of_mass",
        ] and not isinstance(other, type(self)):
            message = "Only center of geometry (COG), 'all', or 'any' is supported "
            message += f"if 'other' is 'Iterable'. '{to}' is reset to 'COG' for 'to'."
            to = "COG"
            warnings.warn(
                message,
                RuntimeWarning,
                stacklevel=2,
            )

        other_data: np.ndarray | tuple | Self = np.asanyarray(other)
        if isinstance(other, type(self)):
            if to.lower() in ["com", "center of mass", "center_of_mass"]:
                to = "COM"
                other_data = np.asanyarray(other.center_of_mass)
                other_data = (other_data[0], other_data[1], other_data[2])
            elif to.lower() in ["cog", "center of geometry", "center_of_geometry"]:
                to = "COG"
                other_data = np.asanyarray(other.center_of_geometry)
                other_data = (other_data[0], other_data[1], other_data[2])
            else:  # 'any' or 'all'
                to = to.upper()
                other_data = other

        else:
            if to.lower() in ["cog", "center of geometry", "center_of_geometry"]:
                to = "COG"
                other_data = np.mean(other, axis=0)
                other_data = (other_data[0], other_data[1], other_data[2])
            else:  # 'any' or 'all'
                to = to.upper()
                other_data = tuple(
                    [
                        (other[i, 0], other[i, 1], other[i, 2])
                        for i in range(other.shape[0])
                    ]
                )

        distance_matrix = PDBDataFrame.get_distance_matrix(
            self,
            other_data=other_data,
        )
        mask = distance_matrix <= cut_off
        if to in ["COG", "COM"]:
            if invert:
                return self.atoms[~mask[:, 0]]
            return self.atoms[mask[:, 0]]

        if (to == "ALL" and not invert) or (to == "ANY" and invert):
            return self.atoms[mask.all(axis=1)]
        return self.atoms[mask.any(axis=1)]

    def _filter_num_col(
        self,
        value: int | float | list[int],
        num_col_name: str,
        relation: str | None = None,
        invert: bool = False,
        epsilon: float | int = 0.01,
        suppress_warning: bool = False,
    ) -> Self:
        """Generic function to do filter by a numerical column.

        Args:
            value (required): value(s) to select by the column given by the
                ``num_col_name`` input.
            num_col_name (required): one of ``atom_number``, ``residue_number``,
                ``x_coord``, ``y_coord``, or ``z_coord``, ``occupancy``, ``b_factor``,
                and  ``nmr_model``. Note: the ``charge`` column is not numerical by
                ``PDB`` format.
            relation (optional): ``x/y/z_coord`` relationship to ``value``.
                It has to be one of ``'<'``, ``'<='``, ``'='``, ``'>='``, and ``'>'``.
                If ``None``, ``'<='`` is used.
                Ignored if a list of integers are provided to ``value``.
                Defaults to **None**.
            invert (optional): whether to invert the selection. Defaults to **False**.
            epsilon (optional): atoms ``abs``(``num_col_value`` - ``value``) <=
                ``epsilon`` are selected when ``invert`` = ``False`` and
                ``relation`` = ``'='``. Ignored if a list of integers are provided to
                ``value``. Defaults to **0.01**.
            suppress_warning (optional): whether to suppress warnings.
                Defaults to **False**.

        Returns:
            sub ``PDBDataFrame``

        Raises:
            ValueError: if xyz not in [``atom_number``, ``residue_number``,
                ``x_coord``, ``y_coord``, or ``z_coord``,
                ``occupancy``, ``b_factor``, ``nmr_model``] or ``relation`` not in
                [``'<'``, ``'<='``, ``'='``, ``'>='``, ``'>'``] when
                selecting on float cols.
        """
        allowed_num_col_names = [
            "atom_number",
            "residue_number",
            "x_coord",
            "y_coord",
            "z_coord",
            "occupancy",
            "b_factor",
            "nmr_model",
        ]
        if num_col_name not in allowed_num_col_names:
            message = f"Only '{allowed_num_col_names}' are allowed in 'num_col_name' "
            message += f"but {num_col_name} was put."
            raise ValueError(message)

        allowed_relations = ["<", "<=", "=", ">=", ">"]
        if relation is None and not isinstance(value, list):
            if isinstance(value, float):
                relation = "<="
            elif isinstance(value, int):
                relation = "="
            else:
                message = "Only 'int', 'float', or 'list[int]' are allowed in 'value' "
                message += f"but {type(value)} was put."
                raise ValueError(message)
        elif not isinstance(value, list) and relation not in allowed_relations:
            message = f"Only '{allowed_relations}' are allowed in 'relation' "
            message += f"but {relation} was put."
            raise ValueError(message)
        elif isinstance(value, list):
            for v in value:
                if not isinstance(v, int):
                    message = "Only 'int' is allowed in 'value' if it is a list, "
                    message += f"but {type(v)} was in {value}."
                    raise ValueError(message)
            if relation is not None:
                relation = None
                if not suppress_warning:
                    message = "'relation' is ignored "
                    message += "when a list is provided to 'value'."
                    warnings.warn(
                        message,
                        RuntimeWarning,
                        stacklevel=2,
                    )

        if relation == "<":
            if invert:
                return self.atoms[self.atoms[num_col_name].values >= value]
            return self.atoms[self.atoms[num_col_name].values < value]
        elif relation == "<=":
            if invert:
                return self.atoms[self.atoms[num_col_name].values > value]
            return self.atoms[self.atoms[num_col_name].values <= value]
        elif relation == "=":
            if invert:
                return self.atoms[
                    np.abs(self.atoms[num_col_name].values - value) >= epsilon
                ]
            return self.atoms[np.abs(self.atoms[num_col_name].values - value) < epsilon]
        elif relation == ">=":
            if invert:
                return self.atoms[self.atoms[num_col_name].values < value]
            return self.atoms[self.atoms[num_col_name].values >= value]
        elif relation == ">":
            if invert:
                return self.atoms[self.atoms[num_col_name].values <= value]
            return self.atoms[self.atoms[num_col_name].values > value]

        # relation is None -> a list of numbers
        if invert:
            return self.atoms[~np.isin(self.atoms[num_col_name].values, value)]
        return self.atoms[np.isin(self.atoms[num_col_name].values, value)]

    @classmethod
    def get_residue_list(cls, pdb_df: Self) -> list[tuple]:
        """Gets the list of residues given a ``PDBDataFrame`` object.

        Args:
            pdb_df (required): a ``PDBDataFrame`` object.

        Returns:
            a list of residues as (``chain_id``, ``residue_name``, ``residue_number``).
        """
        all_residues: list[tuple] = []
        for chain, residue_name, residue_number in zip(
            pdb_df["chain_id"], pdb_df["residue_name"], pdb_df["residue_number"]
        ):
            if residue_name not in pdb_df.RESIDUE_CODES:
                continue
            residue = (chain, residue_name, residue_number)
            if len(all_residues) == 0:
                all_residues.append(residue)
            elif residue != all_residues[-1]:
                all_residues.append(residue)
        return all_residues

    @classmethod
    @functools.lru_cache()
    def get_distance_matrix(
        cls,
        pdb_df: Self,
        other_data: Self | tuple | None = None,
        use_r2: bool = True,
        square_form: bool = False,
    ) -> np.ndarray:
        """Calculates the distance matrix given a ``PDBDataFrame`` object and
        (optional) reference data.

        Args:
            pdb_df (required): a ``PDBDataFrame`` object.
            other_data (optional): the coordinates of to calculate the distances
                against. Defaults to **None**.
            use_r2 (optional): whether to use r^2 or r for distance matrix.
                Defaults to **True**.
            square_form (optional): whether to output a square form of the
                density matrix. If two ``PDBDataFrame``s are different or
                ``other_data`` is not a ``PDBDataFrame``, ``square_form`` is ignored.
                Defaults to **False**.


        Returns:
            distance matrix (squared or condensed form)

        Raises:
            ValueError: if ``other_data`` is not of ``PDBDataFrame|tuple|None`` type or
                wrong shape if it is a ``tuple``.
        """
        cols = ["x_coord", "y_coord", "z_coord"]

        if other_data is None or (
            isinstance(other_data, cls) and pdb_df.atoms == other_data.atoms
        ):
            if not square_form:
                if use_r2:
                    return pdist(pdb_df.atoms[cols].values, "sqeuclidean")
                return pdist(pdb_df.atoms[cols].values)
            if use_r2:
                return squareform(pdist(pdb_df.atoms[cols].values, "sqeuclidean"))
            return squareform(pdist(pdb_df.atoms[cols].values))

        elif isinstance(other_data, cls):
            if use_r2:
                return cdist(
                    pdb_df.atoms[cols].values,
                    other_data.atoms[cols].values,
                    "sqeuclidean",
                )
            return cdist(
                pdb_df.atoms[cols].values, other_data.atoms[cols].values, "euclidean"
            )

        elif isinstance(other_data, tuple):
            other_array = np.asanyarray(other_data)
            if other_array.shape != (3,):
                if not (len(other_array.shape) == 2 and other_array.shape[1] == 3):
                    message = "'other_data' expects a shape of (N, 3) or (3,) "
                    message += f"if given a tuple, but {other_array.shape} was given."
                    raise ValueError(message)
            else:
                other_array = other_array.reshape(1, 3)

            if use_r2:
                return cdist(pdb_df.atoms[cols].values, other_array, "sqeuclidean")
            return cdist(pdb_df.atoms[cols].values, other_array, "euclidean")
        raise ValueError("'other_data' type has to be tuple or PDBDataFrame or None.")



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Mol2 format reading.

Read a Tripos ``.mol2`` file into a dictionary of ``pandas DataFrames``. Different
categories like 'ATOM' and 'BOND' are read into different DataFrame objects.

"""
from __future__ import annotations

import os
import warnings
from collections import defaultdict

import pandas as pd  # type: ignore

IMPLEMENTED_MOL2_CATS = ["ATOM", "MOLECULE", "BOND", "HEADER"]
"""MOL2 categories that are currently implemented."""


ATOM_COL_NAMES = (
    "atom_id",  # int
    "atom_name",  # str
    "x",  # float
    "y",  # float
    "z",  # float
    "atom_type",  # str
    "subst_id",  # int, optional
    "subst_name",  # str, optional
    "charge",  # float, optional
    "status_bit",  # str, optional
)
"""MOL2 ``ATOM`` column names."""


BOND_COL_NAMES = (
    "bond_id",  # int
    "origin_atom_id",  # int
    "target_atom_id",  # int
    "bond_type",  # str
    "status_bit",  # str, optional
)
"""MOL2 ``BOND`` column names."""


def read_mol2(
    mol2_file: str | os.PathLike,
    category_names: list | None = None,
) -> dict[str, pd.DataFrame]:
    """Reads a ``.mol2`` file's categories into a ``dict`` of ``Pandas DataFrame`` s.

    Args:
        mol2_file(required): file name for a PDB file.
        category_names (optional): a list of categories as to the ``.mol2`` file format.
            If ``None``, [``'ATOM'``, ``'MOLECULE'``, ``'BOND'``, ``'HEADER'``] is used.
            Defaults to **None**.

    Returns:
        A dict of ``category_name`` as keys(s) and ``pd.DataFrame`` as values.

    Raises:
        NotImplementedError: if ``category_names`` not a subset of
            [``'ATOM'``, ``'MOLECULE'``, ``'BOND'``, ``'HEADER'``]
    """
    data: dict[str, pd.DataFrame] = {}
    if category_names is None:
        category_names = ["ATOM", "MOLECULE", "BOND", "HEADER"]
    for category_name in category_names:
        if category_name not in IMPLEMENTED_MOL2_CATS:
            implemented = ", ".join(IMPLEMENTED_MOL2_CATS)
            raise NotImplementedError(
                f"""Only {implemented} categories are implemented for the MOL2 format.
                Create an issue at https://github.com/Ruibin-Liu/pdbx2df if
                you want the {category_name} category to be implemented.
                """
            )
        data[category_name] = pd.DataFrame()

    category_block_lines: dict[str, list] = defaultdict(list)
    with open(mol2_file, "r", encoding="utf-8") as mol_f:
        line = mol_f.readline()
        while line:
            if line.startswith("#"):
                category_name = "HEADER"
                line = line.lstrip("#").strip()
                if line:
                    category_block_lines[category_name].append(tuple(line.split(": ")))
            elif line.startswith("@<TRIPOS>"):
                category_name = line.strip()[9:]
                if category_name not in category_names:
                    line = mol_f.readline()
                    continue
                line = mol_f.readline()
                while line and line != "\n" and line[0] != "@":
                    category_block_lines[category_name].append(
                        tuple(line.strip().split())
                    )
                    line = mol_f.readline()
                if line and line[0] == "@":
                    continue
            line = mol_f.readline()
    for category_name in category_names:
        if category_name not in category_block_lines:
            warnings.warn(
                f"The required category {category_name} is not in the file {mol2_file}.",
                RuntimeWarning,
                stacklevel=2,
            )
        elif category_name == "ATOM":
            width = len(category_block_lines[category_name][0])
            data[category_name] = pd.DataFrame(
                category_block_lines[category_name],
                columns=ATOM_COL_NAMES[0:width],
            )
            data[category_name] = _set_atom_df_dtypes(data[category_name])
        elif category_name == "BOND":
            width = len(category_block_lines[category_name][0])
            data[category_name] = pd.DataFrame(
                category_block_lines[category_name],
                columns=BOND_COL_NAMES[0:width],
            )
            data[category_name] = _set_bond_df_dtypes(data[category_name])
        elif category_name == "MOLECULE":
            data[category_name] = _get_molecule_df(category_block_lines[category_name])
        elif category_name == "HEADER":
            data[category_name] = _get_header_df(category_block_lines[category_name])

    return data


def _get_header_df(header_lines: list[tuple]) -> pd.DataFrame:
    """Turns the ``HEADER`` lines into a ``Pandas DataFrame``.
    The ``HEADER`` lines are those starting with one or multiple ``#`` symbols.

    Args:
        header_lines (required): a list of tuples corresponding to
            each line's content. Tuples are generate by splitting the lines by ``:``.

    Returns:
        ``Pandas DataFrame`` of The ``HEADER`` category
    """
    header_attrs: dict[str, list[str]] = defaultdict(list)
    n_no_name = 0
    for header_line in header_lines:
        header_line = tuple([i.strip() for i in header_line])
        if len(header_line) == 1:
            header_attrs[f"info_{n_no_name}"] = [header_line[0]]
        elif len(header_line) == 2:
            header_attrs[header_line[0]] = [header_line[1]]
        else:
            message = f"The line {header_line} has > 2 items separated"
            message += " by ':'. pdbx2df uses the first as column name"
            message += " and concatenate the rest as value."
            warnings.warn(
                message,
                RuntimeWarning,
                stacklevel=2,
            )
            value = ";".join(header_line[1:])
            header_attrs[header_line[0]] = [value]

    return pd.DataFrame(header_attrs)


def _get_molecule_df(molecule_lines: list[tuple]) -> pd.DataFrame:
    """Turns the ``MOLECULE`` lines into a ``Pandas DataFrame``.

    Args:
        molecule_lines (required): a list of tuples corresponding to
            each line's content.

    Returns:
        ``Pandas DataFrame`` of The ``MOLECULE`` category
    """
    molecule_attrs: dict[str, list[str] | list[int]] = {}
    line_0 = {"mol_name": [" ".join(molecule_lines[0])]}
    line_1_names = ["num_atoms", "num_bonds", "num_subst", "num_feat", "num_sets"]
    line_1 = {
        name: [int(value)] for name, value in zip(line_1_names, molecule_lines[1])
    }
    line_2 = {"mol_type": [molecule_lines[2][0]]}
    line_3 = {"charge_type": [molecule_lines[3][0]]}
    molecule_attrs = {**line_0, **line_1, **line_2, **line_3}
    if len(molecule_lines) > 4:
        line_4 = {"status_bits": [molecule_lines[4][0]]}
        molecule_attrs = {**molecule_attrs, **line_4}
    if len(molecule_lines) > 5:
        line_5 = {"mol_comment": [molecule_lines[5][0]]}
        molecule_attrs = {**molecule_attrs, **line_5}

    return pd.DataFrame(molecule_attrs)


def _set_atom_df_dtypes(data_df: pd.DataFrame) -> pd.DataFrame:
    """Sets the data types for the ``ATOM`` category.

    Args:
        data_df (required): original ``Pandas DataFrame``
            for the ``ATOM`` category with all strings.

    Returns:
        ``Pandas DataFrame`` of The ``ATOM`` category
    """
    data_df[["atom_id", "x", "y", "z"]] = data_df[["atom_id", "x", "y", "z"]].astype(
        {"atom_id": "int32", "x": "float32", "y": "float32", "z": "float32"}
    )
    if "subst_id" in data_df.columns:
        data_df["subst_id"] = data_df["subst_id"].astype("int32")
    if "charge" in data_df.columns:
        data_df["charge"] = data_df["charge"].astype("float32")

    return data_df


def _set_bond_df_dtypes(data_df: pd.DataFrame) -> pd.DataFrame:
    """Sets the data types for the ``BOND`` category

    Args:
        data_df (required): original ``Pandas DataFrame``
            for the ``BOND`` category with all strings.

    Returns:
        ``Pandas DataFrame`` of The ``BOND`` category
    """
    data_df[["bond_id", "origin_atom_id", "target_atom_id"]] = data_df[
        ["bond_id", "origin_atom_id", "target_atom_id"]
    ].astype({"bond_id": "int32", "origin_atom_id": "int32", "target_atom_id": "int32"})

    return data_df



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Tests for the PDBDataFrame class."""
import os
import sys

import numpy as np  # type: ignore
import pytest
from scipy.spatial.transform import Rotation  # type: ignore

from pdbx2df.constants import ELEMENT_MASSES
from pdbx2df.pdb_dataframe import RESIDUE_CODES, PDBDataFrame
from pdbx2df.read_pdb import read_pdb

sys.path.append("..")
CFD = os.path.dirname(__file__)


def test_construct_from_df():
    """Testing constructing from plain DataFrame."""
    file_path = [CFD, "test_files", "1G03.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)
    assert isinstance(pdb_df, PDBDataFrame), "Does not return a 'PDBDataFrame instance."
    assert (
        df.columns == pdb_df.columns
    ).all(), "Columns changed when constructing from a plain DataFrame."
    assert (
        df.shape == pdb_df.shape
    ), "DataFrame shape changed when constructing from a plain DataFrame."


def test_hash_eq():
    """Testing whether PDBDataFrame is hashable and comparable."""
    file_path = [CFD, "test_files", "1G03.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)
    with pytest.raises(TypeError) as exception_info:
        assert df == pdb_df
    assert (
        str(exception_info.value) == "'NoneType' object is not callable"
    ), "Compared to non-directly-comparable plain DataFrame."

    pdb_df2 = pdb_df.copy()
    assert pdb_df == pdb_df2, "Can not compare with same type."


def test_property_getter():
    """Testing getting properties."""
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)

    assert (
        len(list(pdb_df.RESIDUE_CODES.keys())[0]) == 4
    ), "Residue name length not 4 for Chimera PDBs."

    assert (
        len(list(pdb_df.ELEMENT_MASSES.keys())[0]) == 2
    ), "Element symbol length not 2 for Chimera PDBs."

    assert pdb_df.is_chimera, "'is_chimera' not correct."

    assert pdb_df.hash_random_state == 0, "Default hash random state is not 0."

    assert (
        pdb_df.use_squared_distance
    ), "Default not using R2 for distance calculations."

    assert (
        not pdb_df.use_square_form
    ), "Default not using condensed form for distance matrix."

    atoms = pdb_df[pdb_df.record_name.isin(["ATOM  ", "HETATM"])]
    assert pdb_df.atoms.shape == atoms.shape, "TER lines not removed correctly."

    coords = pdb_df.atoms[["x_coord", "y_coord", "z_coord"]]
    assert np.allclose(
        coords.values, pdb_df.coords.values
    ), "Coords not selected correctly."
    sequences = {
        "A": "AWEIPRESLRLEVKLGQGCFGEVWMGTWNGTTRVAIKTLKPGTMSPEAFLQEAQVMKKLRHEKLVQLYAVVSEEPIYIVTEYMSKGSLLDFLKGEMGKYLRLPQLVDMAAQIASGMAYVERMNYVHRDLRAANILVGENLVCKVADFGFPIKWTAPEAALYGRFTIKSDVWSFGILLTELTTKGRVPYPGMVNREVLDQVERGYRMPCPPECPESLHDLMCQCWRKDPEERPTFEYLQAFLEDYFTSTEPQYQPGENL",  # noqa
        "B": "AWEIPRESLRLEVKLGQGCFGEVWMGTWNGTTRVAIKTLKPGTMSPEAFLQEAQVMKKLRHEKLVQLYAVVSEEPIYIVTEYMSKGSLLDFLKGEMGKYLRLPQLVDMAAQIASGMAYVERMNYVHRDLRAANILVGENLVCKVADFGLARLFPIKWTAPEAALYGRFTIKSDVWSFGILLTELTTKGRVPYPGMVNREVLDQVERGYRMPCPPECPESLHDLMCQCWRKDPEERPTFEYLQAFLEDYFTSTEPQYQPGENL",  # noqa
    }

    assert pdb_df.sequences == sequences, "Chain sequences not right."

    assert len(pdb_df.residue_list) == 520 and pdb_df.residue_list[0] == (
        "A",
        "ALA ",
        259,
    ), "Residue list not recovered."

    backbone = pdb_df[
        (pdb_df.record_name == "ATOM  ")
        & (pdb_df.atom_name.isin([" C  ", " CA ", " N  ", " O  "]))
        & (pdb_df.element_symbol.isin([" C", " N", " O"]))
    ]
    assert pdb_df.backbone == backbone, "Backbone not correct."

    side_chain = pdb_df[
        (pdb_df.record_name == "ATOM  ")
        & (~pdb_df.atom_name.isin([" C  ", " CA ", " N  ", " O  "]))
        & (pdb_df.element_symbol.isin([" C", " N", " O", " S"]))
    ]
    assert pdb_df.side_chain == side_chain, "Side chain not correct."

    ca_atoms = pdb_df[
        (pdb_df.record_name == "ATOM  ")
        & (pdb_df.atom_name.isin([" CA "]))
        & (pdb_df.element_symbol.isin([" C"]))
    ]
    assert pdb_df.ca_atoms == ca_atoms, "C alpha atoms not correct."

    heavy_atoms = pdb_df[(~pdb_df.element_symbol.isin([" H", " D", " T"]))]
    assert pdb_df.heavy_atoms == heavy_atoms, "Heavy atoms not correct."

    hetero_atoms = pdb_df[(pdb_df.record_name == "HETATM")]
    assert pdb_df.hetero_atoms == hetero_atoms, "Hetero atoms not correct."

    residues = pdb_df[(pdb_df.record_name == "ATOM  ")]
    assert pdb_df.residues == residues, "Residue atoms not correct."

    water = pdb_df[(pdb_df.record_name == "HETATM") & (pdb_df.residue_name == "HOH ")]
    assert pdb_df.water == water, "Water atoms not correct."

    assert pdb_df.n_atoms == len(pdb_df.atoms), "Number of atoms not correct."

    assert pdb_df.n_chains == 2, "Number of chains not correct."

    assert pdb_df.n_segments == 1, "Number of segments not correct."

    assert pdb_df.n_models == 1, "Number of models not correct."
    assert np.allclose(
        pdb_df.center_of_geometry, [4.1473727, 15.076957, 10.4202175]
    ), "Center of geometry not correct."

    assert np.allclose(
        pdb_df.center_of_mass, [4.1451645, 15.094206, 10.414299]
    ), "Center of mass not correct."

    assert (
        abs(pdb_df.radius_of_gyration - 741.9655564097923) <= 0.0000001
    ), "Center of mass not correct."

    assert np.allclose(
        pdb_df.atom_numbers([1, 2, 3]).distance_matrix,
        [2.194553, 6.47993342, 2.34410508],
    ), "Distance matrix not correct."


def test_property_setter():
    """Testing setting properties."""
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)

    pdb_df.RESIDUE_CODES = {"Ala": "A"}
    assert pdb_df.RESIDUE_CODES == {"ALA ": "A"}, "RESIDUE CODES not set correctly."
    pdb_df.RESIDUE_CODES = RESIDUE_CODES

    pdb_df.ELEMENT_MASSES = {"Ca": 1000}
    assert pdb_df.ELEMENT_MASSES == {
        "CA": 1000
    }, "ELEMENT_MASSES CODES not set correctly."
    pdb_df.ELEMENT_MASSES = ELEMENT_MASSES

    pdb_df.hash_random_state = 42
    assert pdb_df.hash_random_state == 42, "hash_random_state not set correctly."

    pdb_df.use_squared_distance = False
    assert not pdb_df.use_squared_distance, "use_squared_distance not set correctly."

    pdb_df.use_square_form = True
    assert pdb_df.use_square_form, "use_square_form not set correctly."


def test_get_residue_list():
    """Testing the get_residue_list class method."""
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)
    residue_list = PDBDataFrame.get_residue_list(pdb_df)

    assert len(residue_list) == 520, "get_residue_list gets length wrong."

    assert residue_list[0] == (
        "A",
        "ALA ",
        259,
    ), "get_residue_list gets first residue wrong."


def test_get_masses():
    """Testing the get_masses method."""
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)
    masses = pdb_df.atoms.get_masses()

    assert masses.shape == (4347,), "get_masses gets length wrong."

    assert np.isclose(masses[0], 14.007), "get_masses gets first atom mass wrong."


def test_get_distance_matrix():
    """Testing the get_distance_matrix class method."""
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)

    # Defaults: use R2 and use condensed matrix
    condensed_dis_matrix = PDBDataFrame.get_distance_matrix(pdb_df)
    assert condensed_dis_matrix.shape == (
        9446031,
    ), "get_distance_matrix gets condensed matrix shape wrong for self distance matrix."
    message = "get_distance_matrix gets 1st element wrong "
    message += "for condensed matrix for self distance matrix."
    assert np.isclose(condensed_dis_matrix[0], 2.194553), message

    # Use R and square matrix
    dis_matrix = PDBDataFrame.get_distance_matrix(
        pdb_df, use_r2=False, square_form=True
    )
    assert dis_matrix.shape == (
        4347,
        4347,
    ), "get_distance_matrix gets square matrix shape wrong for self distance matrix."
    message = "get_distance_matrix gets element [0, 1] wrong for square matrix "
    message += "while using r for self distance matrix."
    assert np.isclose(dis_matrix[0, 1], 1.48140238), message

    # Use 1 point in other_data
    dis_matrix = PDBDataFrame.get_distance_matrix(pdb_df, other_data=(0, 0, 0))
    assert dis_matrix.shape == (
        4347,
        1,
    ), "get_distance_matrix gets matrix shape wrong for distance matrix to one point."
    message = "get_distance_matrix gets element [0, 0] wrong for distance matrix "
    message += "while using r for distance matrix to one point."
    assert np.isclose(dis_matrix[0], 753.49101945), message

    # Use multiple point in other_data
    dis_matrix = PDBDataFrame.get_distance_matrix(
        pdb_df,
        other_data=((0, 0, 0), (1, 1, 1)),
    )
    assert dis_matrix.shape == (
        4347,
        2,
    ), "get_distance_matrix gets matrix shape wrong for distance matrix to two points."
    message = "get_distance_matrix gets element [0, 0] wrong for distance matrix "
    message += "while using r for distance matrix to two points."
    assert np.isclose(dis_matrix[0, 0], 753.49101945), message

    # Use another PDBDataFrame
    dis_matrix = PDBDataFrame.get_distance_matrix(
        pdb_df,
        pdb_df.ca_atoms,
    )
    message = "get_distance_matrix gets matrix shape wrong for distance matrix "
    message += "to another PDBDataFrame."
    assert dis_matrix.shape == (4347, 522), message
    message = "get_distance_matrix gets element [0, 0] wrong for distance matrix "
    message += "while using r for distance matrix to another PDBDataFrame."
    assert np.isclose(dis_matrix[0, 0], 2.194553), message


def test_rmsd():
    """Testing the rmsd method."""
    file_path = [CFD, "test_files", "1G03.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)
    m1 = pdb_df.nmr_models(1).coords.values
    m2 = pdb_df.nmr_models(20).coords.values
    expected_raw = np.sqrt(np.sum((m1 - m2) ** 2) / len(m1))
    assert np.isclose(
        expected_raw, pdb_df.rmsd(align=False)[-1]
    ), "Unaligned RMSD between model 1 and 20 are different."

    m1 = m1 - np.mean(m1, axis=0)
    m2 = m2 - np.mean(m2, axis=0)
    rot = Rotation.align_vectors(m1, m2)[0]
    m2 = rot.apply(m2)
    expected_aligned = np.sqrt(np.sum((m1 - m2) ** 2) / len(m1))
    assert np.isclose(
        expected_aligned, pdb_df.rmsd(align=True)[-1]
    ), "Aligned RMSD between model 1 and 20 are different."


def test_filter_num_col():
    """Testing the _filter_num_col method."""
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)

    # One int
    sub_df = pdb_df._filter_num_col(1, "atom_number")
    expected = pdb_df[pdb_df.atom_number.isin([1])]
    assert sub_df == expected, "Filter by 1 integer failed with default keywords."

    # One float
    sub_df = pdb_df._filter_num_col(1, "x_coord")
    expected = pdb_df[np.abs(pdb_df.x_coord.values - 1) <= 0.01]
    assert sub_df == expected, "Filter by 1 float failed with default keywords."

    # A list of ints
    sub_df = pdb_df._filter_num_col([1, 2], "atom_number")
    expected = pdb_df[pdb_df.atom_number.isin([1, 2])]
    assert (
        sub_df == expected
    ), "Filter by a list of two integer failed with default keywords."

    # Mixed types in list
    except_message = "Only 'int' is allowed in 'value' if it is a list, "
    except_message += "but <class 'float'> was in [1, 2.0]."
    with pytest.raises(ValueError) as exception_info:
        sub_df = pdb_df._filter_num_col([1, 2.0], "atom_number")
    assert (
        str(exception_info.value) == except_message
    ), "Did not catch the error of using mixed types in list."

    # One str
    except_message = "Only 'int', 'float', or 'list[int]' are allowed in 'value' "
    except_message += "but <class 'str'> was put."
    with pytest.raises(ValueError) as exception_info:
        sub_df = pdb_df._filter_num_col("1", "atom_number")
    assert (
        str(exception_info.value) == except_message
    ), "Did not catch the error of using a str type."

    # Allowed columns
    except_message = "Only '['atom_number', 'residue_number', 'x_coord', 'y_coord', "
    except_message += "'z_coord', 'occupancy', 'b_factor', 'nmr_model']' are allowed "
    except_message += "in 'num_col_name' but atom_name was put."
    with pytest.raises(ValueError) as exception_info:
        sub_df = pdb_df._filter_num_col(1, "atom_name")
    assert (
        str(exception_info.value) == except_message
    ), "Did not catch the error of using a wrong column name."

    # Wrong relations
    except_message = "Only '['<', '<=', '=', '>=', '>']' are allowed "
    except_message += "in 'relation' but ? was put."
    with pytest.raises(ValueError) as exception_info:
        sub_df = pdb_df._filter_num_col(1, "atom_number", relation="?")
    assert (
        str(exception_info.value) == except_message
    ), "Did not catch the error of using a wrong 'relation'."

    # relation for list
    warning_message = "'relation' is ignored when a list is provided to 'value'."
    with pytest.warns(RuntimeWarning) as warning_info:
        sub_df = pdb_df._filter_num_col([1, 2], "atom_number", relation="?")
    assert (
        warning_info[0].message.args[0] == warning_message
    ), "Did not catch the warning of using 'relation' when a list is given."

    # relation and invert
    sub_df_1 = pdb_df._filter_num_col(2, "atom_number", relation="<=", invert=False)
    sub_df_2 = pdb_df._filter_num_col(2, "atom_number", relation=">", invert=True)
    assert (
        sub_df_1 == sub_df_2
    ), "<= and invert=False should have been the same as > and invert=True"


def test_atom_selections():
    """Testing atom selection functions."""
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)

    # record_name
    sub_df = pdb_df.record_names(["HETATM"])
    expected = pdb_df[pdb_df.record_name == "HETATM"]
    assert sub_df == expected, "record_names selection failed."

    # atom_number
    sub_df = pdb_df.atom_numbers([1, 2, 3])
    expected = pdb_df[pdb_df.atom_number.isin([1, 2, 3])]
    assert sub_df == expected, "atom_numbers selection failed."

    # atom_name
    sub_df = pdb_df.atom_names(["CA", "C"])
    expected = pdb_df[pdb_df.atom_name.isin([" CA ", " C  "])]
    assert sub_df == expected, "atom_names selection failed."

    # alt_loc
    sub_df = pdb_df.alt_locs(["B"])
    expected = pdb_df[pdb_df.alt_loc.isin(["B", " "])]
    assert (
        sub_df.alt_loc.unique() == expected.alt_loc.unique()
    ).all(), "alt_locs selection failed."

    # residue_name
    sub_df = pdb_df.residue_names(["ALA"])
    expected = pdb_df[pdb_df.residue_name.isin(["ALA "])]
    assert sub_df == expected, "residue_names selection failed."

    # chain_id
    sub_df = pdb_df.chain_ids(["A"])
    expected = pdb_df[pdb_df.chain_id.isin(["A"])]
    assert sub_df == expected, "chain_ids selection failed."

    # residue_number
    sub_df = pdb_df.residue_numbers([300, 301])
    expected = pdb_df[pdb_df.residue_number.isin([300, 301])]
    assert sub_df == expected, "residue_numbers selection failed."

    # insertion
    sub_df = pdb_df.insertions([" "])
    expected = pdb_df[pdb_df.insertion.isin([" "])]
    assert sub_df == expected, "insertions selection failed with insertion=' '."
    assert (
        len(pdb_df.insertions([""])) == 0
    ), "insertions selection failed with insertion=''."

    # x_coord
    sub_df = pdb_df.x_coords(1.0)
    expected = pdb_df[pdb_df.x_coord.values <= 1.0]
    assert sub_df == expected, "x_coords selection failed."

    # y_coord
    sub_df = pdb_df.y_coords(1.0, relation=">")
    expected = pdb_df[pdb_df.y_coord.values > 1.0]
    assert sub_df == expected, "y_coords selection failed."

    # z_coord
    sub_df = pdb_df.z_coords(1.0, relation="=", epsilon=1.0)
    expected = pdb_df[np.abs(pdb_df.z_coord.values - 1.0) <= 1.0]
    assert sub_df == expected, "z_coords selection failed."

    # occupancy
    sub_df = pdb_df.occupancies(1.0, relation="=", epsilon=0.0001)
    expected = pdb_df[np.abs(pdb_df.occupancy.values - 1.0) <= 0.0001]
    assert sub_df == expected, "occupancies selection failed."

    # b_factor
    sub_df = pdb_df.b_factors(50, relation=">=")
    expected = pdb_df[pdb_df.b_factor.values >= 50]
    assert sub_df == expected, "b_factors selection failed."

    # segment_id
    sub_df = pdb_df.segment_ids(["    "])
    expected = pdb_df
    assert sub_df == expected, "segment_ids selection failed."

    # element_symbol
    sub_df = pdb_df.element_symbols(["O", "S"])
    expected = pdb_df[pdb_df.element_symbol.isin([" O", " S"])]
    assert sub_df == expected, "element_symbols selection failed."

    # charge
    sub_df = pdb_df.charges(["  "])
    expected = pdb_df
    assert sub_df == expected, "charges selection failed."

    # distances
    first_atom = pdb_df.head(1)
    sub_df = pdb_df.distances(first_atom, cut_off=1.49, to="all")
    expected = pdb_df.atom_numbers([1, 2])
    assert (
        sub_df == expected
    ), "distances selection failed with a PDBDataFrame instance."

    sub_df = pdb_df.distances([-1.972, 3.107, -27.202], cut_off=1.49, to="all")
    expected = pdb_df.atom_numbers([1, 2])
    assert sub_df == expected, "distances selection failed with one point."

    points = [
        [-1.972, 3.107, -27.202],
        [-0.601, 3.663, -27.278],
    ]
    sub_df = pdb_df.distances(points, cut_off=1.49, to="any")
    expected = pdb_df.atom_numbers([1, 2])
    assert sub_df == expected, "distances selection failed with two points."

    sub_df_1 = pdb_df.distances(points, cut_off=0, to="any", invert=True)
    sub_df_2 = pdb_df.distances(points, cut_off=0, to="all", invert=False)
    assert (
        sub_df_2 == sub_df_1
    ), "to='any' and invert=True not the same as to='all' and invert=False."


def test_nmr_models():
    """Testing atom selection using nmr_model."""
    file_path = [CFD, "test_files", "1G03.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)

    # nmr_model
    sub_df = pdb_df.nmr_models([1, 2, 20])
    expected = pdb_df.atoms[pdb_df.atoms.nmr_model.isin([1, 2, 20])]
    assert sub_df == expected, "nmr_models selection failed."
    assert (
        sub_df.nmr_model.unique() == [1, 2, 20]
    ).all(), "Not all 1,2&20 models are selected."

    # Without nmr_model column
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    df = pdb["_atom_site"]
    pdb_df = PDBDataFrame(df)
    sub_df = pdb_df.nmr_models([1, 2, 20])
    expected = pdb_df
    assert sub_df == expected, "nmr_models selection failed when there is only 1 model."



# Configuration file for the Sphinx documentation builder.
import os
import sys

sys.path.insert(0, os.path.abspath("../"))

from pdbx2df.version import __version__ as pdbx2df_version  # type: ignore

# -- Project information

project = "pdbx2df"
copyright = "2023, Ruibin Liu"
author = "Ruibin Liu"

version = pdbx2df_version
release = pdbx2df_version

# -- General configuration

extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
]

napoleon_google_docstring = True

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

autodoc_member_order = "bysource"

# -- Options for HTML output

html_theme = "furo"
# html_theme = "sphinx_rtd_theme"

# -- Options for EPUB output
# epub_show_urls = 'footnote'



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Split a line in mmCIF files."""
from __future__ import annotations


def split_line(line: str, delimiter: str | None = None) -> list:
    """Splits a string line into tokens separated by ``delimiter`` s, assuming all
    ``'`` and ``"`` in the start character or following a `delimiter` are paired to
    quote a token.

    Args:
        line (required): line as a string
        delimiter (optional): ``delimiter`` to split the line.
            If ``None``, ``' '`` (one space) is used.
            Defaults to **None**.

    Returns:
        A list of tokens
    """
    if not delimiter:
        delimiter = " "
    words = []
    # without quotes, using shlex
    if '"' not in line and "'" not in line:
        if delimiter == " ":
            return line.split()
        return line.split(delimiter)

    # with quotes
    single_start = False
    double_start = False
    tmp: list[str] = []
    for i, char in enumerate(line):
        # quoted by single quotes ''
        if i == 0 and char == "'":  # line starting with '
            single_start = True
        elif (
            char == "'"
            and line[i - 1] == delimiter
            and not double_start
            and not single_start
        ):  # a new part quoted with '
            single_start = True
        elif char == "'" and single_start:  # a part quoted with ' ended
            single_start = False  # reset
        # quoted by double quotes ""
        elif i == 0 and char == '"':  # line starting with "
            double_start = True
        elif (
            char == '"'
            and line[i - 1] == delimiter
            and not double_start
            and not single_start
        ):  # a new part quoted with "
            double_start = True
        elif char == '"' and double_start:  # a part quoted with " ended
            double_start = False  # reset
        elif (
            char in [delimiter, "\n", "\r"] and not single_start and not double_start
        ):  # a part not quoted ended
            if tmp:
                words.append("".join(tmp))
                tmp = []
        else:  # Other characters including space in quotes
            tmp.append(char)
        if (
            tmp and i == len(line) - 1
        ):  # in case no '\r', '\n', or delimiter is at the end
            words.append("".join(tmp))
    if single_start or double_start:
        raise ValueError("Bad line: quotes not paired!")

    return words



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Tests for splitting lines in PDBx/mmCIF files."""
import sys

import pytest  # type: ignore

from pdbx2df.split_line import split_line

sys.path.append("..")


def test_split_line():
    """
    Test split_line function
    """
    # Test line start with '  and " quoted by ''
    line = """'a "T"' in Test
    """
    expected = ['a "T"', "in", "Test"]
    assert split_line(line) == expected

    # Test line start with " and ' quoted by ""
    line = """"a 't'" in Test
    """
    expected = ["a 't'", "in", "Test"]
    assert split_line(line) == expected

    # Test '' and "" both used as quotes
    line = """
    DA  'DNA linking'       y "2'-DEOXYADENOSINE-5'-MONOPHOSPHATE" ? 'C10 H14 N5 O6 P' 331.222
    """  # noqa
    expected = [
        "DA",
        "DNA linking",
        "y",
        "2'-DEOXYADENOSINE-5'-MONOPHOSPHATE",
        "?",
        "C10 H14 N5 O6 P",
        "331.222",
    ]
    assert split_line(line) == expected

    # Test ending with '\n'
    line = "Test\n return symbol in middle"
    expected = ["Test", "return", "symbol", "in", "middle"]
    assert split_line(line) == expected

    # Test ending with '\t'
    line = "This is a test\t"
    expected = ["This", "is", "a", "test"]
    assert split_line(line) == expected

    # Test ending with multiple delimiter
    line = "This is a test    "
    expected = ["This", "is", "a", "test"]
    assert split_line(line) == expected

    # Test no ending symbol
    line = "This is a test"
    expected = ["This", "is", "a", "test"]
    assert split_line(line) == expected

    # Test not not paired ' or "
    line = 'This "test fails'
    with pytest.raises(ValueError) as e:
        assert split_line(line)
    assert str(e.value) == "Bad line: quotes not paired!"
    line = "This 'test fails"
    with pytest.raises(ValueError) as e:
        assert split_line(line)
    assert str(e.value) == "Bad line: quotes not paired!"

    # Test , as delimiter
    line = 'This,is,  a,"test"'
    expected = ["This", "is", "  a", "test"]
    assert split_line(line, delimiter=",") == expected



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""PDB format reading.

Reads a PDB file, including Chimera compatible ones, into a dict of
``Pandas DataFrame`` s.

Atom coordinates and sequences are currently supported by the following category names:

    ``_atom_site``:  ``ATOM``, ``HETATM``, ``TER``, ``NUMMDL``, ``MODEL``, and
    ``ENDMDL`` lines.

    ``_seq_res``: ``SEQRES`` lines.

"""
from __future__ import annotations

import io
import os
import urllib.request
import warnings
from collections import defaultdict
from pathlib import Path

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from .constants import AMINO_ACIDS

IMPLEMENTED_PDB_CATS = ["_atom_site", "_seq_res"]
"""PDB categories that are currently implemented."""

ATOM_SITE = ("ATOM", "HETATM", "TER")
"""``_atom_site`` primary lines."""

NMR_MDL = ("NUMMDL", "MODEL", "ENDMDL")
"""``_atom_site`` additional lines."""

AF2_MODEL = 4
"""For AlphaFold structures, the version to use."""


def read_pdb(
    pdb_file: str | os.PathLike[str] | None = None,
    pdb_id: str | None = None,
    category_names: list | None = None,
    save_pdb_file: bool = True,
    pdb_file_dir: str | os.PathLike | None = None,
    allow_chimera: bool = True,
    need_ter_lines: bool = True,
) -> dict[str, pd.DataFrame]:
    """Reads a ``.pdb`` file's categories into a ``dict`` of ``Pandas DataFrame`` s.

    Args:
        pdb_id (optional): PDB/Uniprot ID. Required if ``pdb_file`` is ``None``.
            Defaults to **None**.
        pdb_file (optional): file name for a PDB file. Used over `pdb_id`.
            Defaults to **None**.
        category_names (optional): a list of categories similar to the mmCIF format.
            If ``None``, ``_atom_site`` is used.
            To be consistent with the PDBx format, the following category names are
            used to refer to block(s) in a PDB file and only they are supported:

                1. ``_atom_site``: ``ATOM``, ``HETATM``, and ``TER`` lines and possible
                ``NUMMDL``, ``MODEL``, and ``ENDMDL`` lines.

                2. ``_seq_res``: ``SEQRES`` lines.

            Defaults to **None**.
        save_pdb_file (optional): whether to save the fetched PDB file from RCSB
            to ``pdb_file_dir``. Defaults to **True**.
        pdb_file_dir (optional): directory to save fetched PDB files. If ``None`` but
            ``save_pdb_file`` is ``True``, the current working directory is used.
            Defaults to **None**.
        allow_chimera (optional): whether to allow Chimera-formatted PDB files.
            Defaults to **True**.
        need_ter_lines (optional): whether to read the ``TER`` lines into the
            ``DataFrame``. Defaults to **True**.

    Returns:
        A dict of ``Pandas DataFrame`` s corresponding to required categories.

    Raises:
        ValueError: if none of ``pdb_id`` or ``pdbx_file`` is provided, or if ``pdb_id``
            is given but cannot the PDB file cannot be downloaded from RCSB,
        NotImplementedError: if `category_names` not a subset of allowed names.
        FileNotFoundError: if ``pdbx_file`` cannot be found.
    """
    data: dict[str, pd.DataFrame] = {}
    if pdb_id is None and pdb_file is None:
        raise ValueError("At least one of pdb_id and pdb_file has to be given.")
    elif pdb_file is None:
        pdb_id = str(pdb_id).upper()
        if len(pdb_id) == 4:
            pdb_file_url = f"https://files.rcsb.org/view/{pdb_id.upper()}.pdb"
        else:
            pdb_file_url = f"https://alphafold.ebi.ac.uk/files/AF-{pdb_id.upper()}-F1-model_v{AF2_MODEL}.pdb"
        try:
            with urllib.request.urlopen(pdb_file_url) as response:
                raw_data = response.read()
            text = raw_data.decode("utf-8")
            pdb_file_handle: io.TextIOWrapper | io.StringIO = io.StringIO(text)
            if save_pdb_file:
                if pdb_file_dir is None:
                    pdb_file_dir = "./PDB_files"
                pdb_file_dir = Path(pdb_file_dir)
                if not pdb_file_dir.exists():
                    pdb_file_dir.mkdir(parents=True, exist_ok=True)
                file_path = Path(pdb_file_dir, f"{pdb_id}.pdb")
                with open(file_path, "w", encoding="utf-8") as p_file:
                    p_file.write(text)
        except urllib.error.HTTPError as http_error:
            raise ValueError(
                f"Cannot download PDB file from url {pdb_file_url}."
            ) from http_error
    else:
        pdb_file = Path(pdb_file)
        if not pdb_file.exists():
            raise FileNotFoundError(f"File {pdb_file} not found.")
        pdb_file_handle = open(pdb_file, "r", encoding="utf-8")
    if category_names is None:
        category_names = ["_atom_site"]
    for category_name in category_names:
        if category_name not in IMPLEMENTED_PDB_CATS:
            implemented = ", ".join(IMPLEMENTED_PDB_CATS)
            raise NotImplementedError(
                f"Only {implemented} are implemented for the PDB format."
            )
        data[category_name] = pd.DataFrame()

    num_nmr_models = 0
    n_model_lines = 0
    nmr_model = -1  # -1 means there is no NMR model
    n_endmdl_lines = 0
    # atom_site_rows = []
    column_name_types = [
        ("record_name", "U6"),
        ("atom_number", "i4"),
        ("atom_name", "U4"),
        ("alt_loc", "U1"),
        ("residue_name", "U4"),
        ("chain_id", "U1"),
        ("residue_number", "i2"),
        ("insertion", "U1"),
        ("x_coord", "f4"),
        ("y_coord", "f4"),
        ("z_coord", "f4"),
        ("occupancy", "f4"),
        ("b_factor", "f4"),
        ("segment_id", "U4"),
        ("element_symbol", "U2"),
        ("charge", "U2"),
        ("nmr_model", "i2"),
    ]

    if not need_ter_lines:
        all_atom_site_reads = ("ATOM", "HETATM") + NMR_MDL
    else:
        all_atom_site_reads = ATOM_SITE + NMR_MDL  # type: ignore

    if pdb_file is not None:  # This check is just for mypy
        file_stat = os.stat(pdb_file)
        total_lines = int(file_stat.st_size / 81)
    else:
        total_lines = 100000
    n_record = 0
    n_lines_till_atom_lines = 0
    with pdb_file_handle:
        line = pdb_file_handle.readline()
        for line in pdb_file_handle:
            if line.startswith(all_atom_site_reads) and "_atom_site" in category_names:
                if (not n_record) and line.startswith(("ATOM", "HETATM", "MODEL")):
                    array = np.zeros(
                        total_lines - n_lines_till_atom_lines, column_name_types
                    )
                if line.startswith(("ATOM", "HETATM")):
                    array[n_record] = _split_atom_line(
                        line,
                        nmr_model=nmr_model,
                        allow_chimera=allow_chimera,
                        is_ter_line=False,
                    )
                    n_record += 1
                elif need_ter_lines and line.startswith("TER"):
                    array[n_record] = _split_atom_line(
                        line,
                        nmr_model=nmr_model,
                        allow_chimera=allow_chimera,
                        is_ter_line=True,
                    )
                    n_record += 1
                elif line.startswith("MODEL"):
                    n_model_lines += 1
                    nmr_model = int(line[6:].strip())
                elif line.startswith("ENDMDL"):
                    n_endmdl_lines += 1
                    if num_nmr_models and n_endmdl_lines == num_nmr_models:
                        break
                else:  # line.startswith("NUMMDL")
                    num_nmr_models = int(line[6:].strip())
            elif line.startswith("SEQRES") and "_seq_res" in category_names:
                chain_residues: dict[str, list[str]] = defaultdict(list)
                chain_lengths: dict[str, int] = defaultdict(int)
                while line.startswith("SEQRES"):
                    items: list[str] = line.strip().split()
                    chain_id: str = items[2]
                    chain_length: int = int(items[3])
                    chain_lengths[chain_id] = chain_length
                    chain_residues[chain_id].extend(items[4:])
                    if not n_record and "_atom_site" in category_names:
                        n_lines_till_atom_lines += 1
                    line = pdb_file_handle.readline()

            if not n_record and "_atom_site" in category_names:
                n_lines_till_atom_lines += 1

    if (
        "_atom_site" in category_names
        and num_nmr_models != n_model_lines
        and num_nmr_models != 0
    ):
        warnings.warn(
            f"The NUMMDL says {num_nmr_models} NMR models, but {n_model_lines} found.",
            RuntimeWarning,
            stacklevel=2,
        )

    if "_atom_site" in category_names:
        df_atom_site = pd.DataFrame.from_records(array[:n_record])
        data["_atom_site"] = df_atom_site
        if not n_model_lines:
            data["_atom_site"].drop(columns=["nmr_model"], inplace=True)

    if "_seq_res" in category_names:
        chain_sequences: dict[str, str] = {}
        for chain_id, residues in chain_residues.items():
            chain_sequences[chain_id] = "".join([AMINO_ACIDS[i] for i in residues])
            if len(residues) != chain_lengths[chain_id]:
                warnings.warn(
                    f"Cols 14-17 value {chain_lengths[chain_id]} contrasts real length {len(residues)}.",  # noqa
                    RuntimeWarning,
                    stacklevel=2,
                )
        chain_ids = chain_sequences.keys()
        chain_num_residues = [len(chain_sequences[chain_id]) for chain_id in chain_ids]
        chain_seqs = [chain_sequences[chain_id] for chain_id in chain_ids]
        data["_seq_res"] = pd.DataFrame(
            {
                "chain_id": chain_ids,
                "chain_sequence": chain_seqs,
                "chain_length": chain_num_residues,
            }
        )

    return data


def _split_atom_line(
    line: str,
    nmr_model: int = -1,
    allow_chimera: bool = True,
    is_ter_line: bool = False,
) -> tuple:
    """Internal function to parse a single line belonging to ``ATOM``, ``HETATM``,
    or ``TER`` lines.

    Args:
        line (required): A ``ATOM``, ``HETATM``, or ``TER`` line.
        nmr_model (optional): the NMR model number for the line; ``-1`` means
            not an NMR model. Defaults to **-1**.
        allow_chimera (optional): try to parse as a Chimera-formatted PDB file.
            Defaults to **True**
        is_ter_line (optional): whether the line starts with ``TER``.
            Defaults to **False**.

    Returns:
        parsed values as a tuple.
    """
    if not (is_ter_line or allow_chimera):
        return (
            line[0:6],
            line[6:11],
            line[12:16],
            line[16],
            line[17:20],
            line[21],
            line[22:26],
            line[26],
            line[30:38],
            line[38:46],
            line[46:54],
            line[54:60],
            line[60:66],
            line[72:76],
            line[76:78],
            line[78:80],
            nmr_model,
        )
    elif is_ter_line:
        if allow_chimera:
            record_name = line[0:5] + " "
            residue_name = line[17:21]
        else:
            record_name = line[0:6]
            residue_name = line[17:20]
        return (
            record_name,
            line[6:11],
            line[12:16],
            line[16],
            residue_name,
            line[21],
            line[22:26],
            line[26],
            None,
            None,
            None,
            None,
            None,
            line[72:76],
            line[76:78],
            line[78:80],
            nmr_model,
        )
    else:
        if line[0:5] == "HETAT":
            record_name = "HETATM"
            atom_number = line[6:11]
        else:
            record_name = line[0:5] + " "
            atom_number = line[5:11]
        return (
            record_name,
            atom_number,
            line[12:16],
            line[16],
            line[17:21],
            line[21],
            line[22:26],
            line[26],
            line[30:38],
            line[38:46],
            line[46:54],
            line[54:60],
            line[60:66],
            line[72:76],
            line[76:78],
            line[78:80],
            nmr_model,
        )



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Tests for reading PDB files."""
import os
import sys

from pdbx2df.read_pdb import read_pdb

sys.path.append("..")
CFD = os.path.dirname(__file__)


def test_read_pdb():
    """
    Test read_pdb function
    """
    # NMR PDB
    file_path = [CFD, "test_files", "1G03.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    pdb_df = pdb["_atom_site"]
    assert (pdb_df.tail(1).nmr_model == 20).item(), "NMR model index not read as 20."
    assert (
        pdb_df.tail(1).record_name == "TER   "
    ).item(), "Last record name not 'TER   ' when TER lines are required."
    pdb = read_pdb(
        pdb_file=test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=False,
    )
    pdb_df = pdb["_atom_site"]
    assert (
        pdb_df.tail(1).record_name == "ATOM  "
    ).item(), "Last record name not 'ATOM  ' when TER lines are not required."

    # From PDB ID
    pdb_id = "1VIi"
    pdb = read_pdb(
        pdb_id=pdb_id,
        category_names=["_atom_site"],
        save_pdb_file=True,
        allow_chimera=False,
    )
    pdb_df = pdb["_atom_site"]
    assert os.path.exists(
        f"./PDB_files/{pdb_id.upper()}.pdb"
    ), "File not saved to PDB_files if pdb_file_dir not provided."
    assert (
        pdb_df.head(1).residue_name == "MET"
    ).item(), "First 1VII residue is not MET."

    # From Uniprot ID
    uniprot_id = "P01116"
    pdb = read_pdb(
        pdb_id=uniprot_id,
        category_names=["_atom_site"],
        save_pdb_file=True,
        allow_chimera=False,
    )
    pdb_df = pdb["_atom_site"]
    assert os.path.exists(
        f"./PDB_files/{uniprot_id.upper()}.pdb"
    ), "File not saved to PDB_files if pdb_file_dir not provided."
    assert (
        pdb_df.head(1).residue_number == 1
    ).item(), "First P01116 residue number is not 1."

    # Non-NMR PDB
    file_path = [CFD, "test_files", "5K9I.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(pdb_file=test_file, category_names=["_atom_site"])
    pdb_df = pdb["_atom_site"]
    assert (
        "nmr_model" not in pdb_df.columns
    ), "nmr_model column appears when the file is not NMR."

    # Chimera
    file_path = [CFD, "test_files", "5K9I_mimic_Chimera.pdb"]
    test_file = f"{os.sep}".join(file_path)
    pdb = read_pdb(
        test_file,
        category_names=["_atom_site"],
        allow_chimera=True,
        need_ter_lines=True,
    )
    pdb_df = pdb["_atom_site"]
    pdb_df = pdb_df[pdb_df.record_name == "TER   "]
    assert (
        pdb_df.tail(1).atom_number == 14209
    ).item(), "Chimera PDB didn't get larger than 9999 atoms."
    assert (
        pdb_df.tail(1).residue_name == "LEUX"
    ).item(), "Chimera PDB didn't get 4-letter residue names."


def test_read_seq_res():
    """Test reading the 'SEQRES' section."""
    pdb_id = "1VII"
    pdb = read_pdb(
        pdb_id=pdb_id,
        category_names=["_seq_res"],
        save_pdb_file=True,
        allow_chimera=False,
    )
    seq_df = pdb["_seq_res"]
    assert seq_df["chain_sequence"].to_list() == [
        "MLSDEDFKAVFGMTRSAFANLPLWKQQNLKKEKGLF"
    ], "1VII sequence not read in correctly."



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""
# pdbx2df - Super lightweight and fast mmCIF/PDB/MOL2 file parser into Pandas DataFrames and backwards writer.

### Key features:

### 1. Read a mmCIF/PDB/MOL2 file as a dictionary of Pandas DataFrame.

### 2. Atom selection using dot syntax enabled by a DataFrame subclass: PDBDataFrame

### 3. Write a dictionary of Pandas DataFrames into mmCIF/PDB files.
"""  # noqa
from pdbx2df.constants import AMINO_ACIDS, ELEMENT_MASSES
from pdbx2df.pdb_dataframe import PDBDataFrame
from pdbx2df.read_mol2 import read_mol2
from pdbx2df.read_pdb import read_pdb
from pdbx2df.read_pdbx import read_pdbx
from pdbx2df.split_line import split_line
from pdbx2df.write_mol2 import write_mol2
from pdbx2df.write_pdb import write_pdb
from pdbx2df.write_pdbx import write_pdbx

from .version import __version__

__all__ = [
    "PDBDataFrame",
    "read_pdbx",
    "write_pdbx",
    "read_pdb",
    "write_pdb",
    "read_mol2",
    "write_mol2",
    "split_line",
    "AMINO_ACIDS",
    "ELEMENT_MASSES",
    "__version__",
]



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""PDBx/mmCIF format writing.

Write a dict of ``Pandas DataFrame`` back to a PDBx file.

"""
from __future__ import annotations

import os
from collections import defaultdict
from pathlib import Path

import pandas as pd  # type: ignore


def write_pdbx(
    pdbx: dict[str, pd.DataFrame], file_name: str | os.PathLike | None = None
) -> None:
    """Writes a ``dict`` of ``Pandas DataFrame`` s into a PDBx file.

    Args:
        pdbx (required): a ``dict`` of ``Pandas DataFrame`` s to write.
        file_name (optional): file name to write a PDBx file. If ``None``,
            ``pdbx2df_output.cif`` will be used as the file name.
            Defaults to **None**.

    Raises:
        TypeError: if ``pdbx`` is not a valid dict of ``DataFrame``.
    """
    if not file_name:
        file_name = "pdbx2df_output.cif"

    if not isinstance(pdbx, dict):
        raise TypeError(f"pdbx has to be a dict but {type(pdbx)} is provided.")
    multi_record: dict[str, int] = defaultdict(bool)
    max_tag_length: dict[str, int] = defaultdict(int)
    for category_name, records in pdbx.items():
        if not isinstance(records, pd.DataFrame):
            raise TypeError(
                f"pdbx values have to be Pandas DataFrames but {category_name} is a {type(records)}."  # noqa
            )
        if len(records) > 1:
            multi_record[category_name] = True
        for col in records.columns:
            tag_length = len(category_name) + 1 + len(col)
            max_tag_length[category_name] = max(
                max_tag_length[category_name], tag_length
            )
    with open(file_name, "w", encoding="utf-8") as f:
        # write header
        target_name = Path(file_name).name
        if ".cif" == target_name[-4:]:
            f.write(f"data_{target_name[:-4]}\n")
        else:
            f.write(f"data_{target_name}\n")
        # write each category
        for category_name, records in pdbx.items():
            # categories that only have a record
            f.write("#\n")
            if not multi_record[category_name]:
                for col in records.columns:
                    tag = f"{category_name}.{col}"
                    f.write(f"{tag:{max_tag_length[category_name]+3}}")
                    content = records[col].iloc[0]
                    if '"' in content and "'" in content:
                        raise ValueError(
                            f"'{content}' cannot be written into a PDBx file."
                        )
                    elif "'" in content:
                        content = f'"{content}"'
                    elif '"' in content:
                        content = f"'{content}'"
                    elif " " in content:
                        content = f"'{content}'"

                    content_length = len(content)
                    if tag_length + content_length > 130:
                        content = content.strip('"').strip("'")
                        f.write("\n;")
                        if category_name == "_struct_ref":
                            for i in range(0, content_length // 80):
                                f.write(f"{content[80*i:80*(i+1)]}\n")
                        else:
                            f.write(f"{content}\n")
                        f.write(";\n")
                    else:
                        f.write(f"{content}\n")

            # categories that have multiple records
            else:
                max_col_length = defaultdict(int)
                for col in records.columns:
                    if records[col].dtype == "int":
                        max_col_length[col] = len(str(max(records[col])))
                    elif col == "occupancy":
                        max_col_length[col] = 4
                    elif records[col].dtype == "float":
                        max_int_width = max(
                            len(str(int(max(records[col])))),
                            len(str(int(min(records[col])))),
                        )
                        max_col_length[col] = max_int_width + 4
                        if col == "B_iso_or_equiv":
                            max_col_length[col] = max_int_width + 3
                    else:
                        max_col_length[col] = max(records[col].str.len())
                        if records[col].str.contains(" ").any():
                            max_col_length[col] = max_col_length[col] + 1

                    f.write(f"{category_name}.{col}\n")
                for _, record in records.iterrows():
                    for col in records.columns:
                        content = record[col]
                        pad_length = max_col_length[col]
                        if isinstance(content, str):
                            if '"' in content and "'" in content:
                                raise ValueError(
                                    f"'{content}' cannot be written into a PDBx file."
                                )
                            elif "'" in content:
                                content = f'"{content}"'
                            elif '"' in content:
                                content = f"'{content}'"
                            elif " " in content:
                                content = f"'{content}'"

                            f.write(f"{content:<{pad_length+1}}")
                        elif isinstance(content, int):
                            f.write(f"{content:<{pad_length+1}}")
                        elif isinstance(content, float) and col in [
                            "Cartn_x",
                            "Cartn_y",
                            "Cartn_z",
                        ]:
                            f.write(f"{content:<{pad_length+1}.3f}")
                        elif isinstance(content, float) and col in [
                            "occupancy",
                            "B_iso_or_equiv",
                        ]:
                            f.write(f"{content:<{pad_length+1}.2f}")
                    f.write("\n")

        f.write("#")



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Tests for writing MOL2 files."""
import os
import sys

from pdbx2df.read_mol2 import read_mol2
from pdbx2df.write_mol2 import write_mol2

sys.path.append("..")
CFD = os.path.dirname(__file__)


def test_write_mol2():
    """
    Test write_mol2 function
    """
    test_file = [CFD, "test_files", "test.mol2"]
    test_file = f"{os.sep}".join(test_file)
    mol2 = read_mol2(mol2_file=test_file)

    compare_to = [CFD, "test_files", "test_pdbx2df.mol2"]
    compare_to = f"{os.sep}".join(compare_to)

    write_to = [CFD, "test_files", "test_test.mol2"]
    write_to = f"{os.sep}".join(write_to)

    write_mol2(mol2, file_name=write_to)
    with open(compare_to, "r", encoding="utf-8") as cf, open(
        write_to,
        "r",
        encoding="utf-8",
    ) as wf:
        for _ in range(4):
            next(cf)
            next(wf)
        cf_lines = cf.readlines()
        wf_lines = wf.readlines()
        assert cf_lines == wf_lines
    os.remove(write_to)



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Tests for writing PDB files."""
import os
import sys

from pdbx2df.read_pdb import read_pdb
from pdbx2df.write_pdb import write_pdb

sys.path.append("..")
CFD = os.path.dirname(__file__)


def test_write_pdb():
    """
    Test write_pdb function
    """
    pdb_id = "1G03"
    test_file = [CFD, "test_files", f"{pdb_id}.pdb"]
    test_file = f"{os.sep}".join(test_file)
    pdb = read_pdb(
        pdb_file=test_file, category_names=["_atom_site"], allow_chimera=False
    )

    compare_to = [CFD, "test_files", f"{pdb_id}_pdbx2df.pdb"]
    compare_to = f"{os.sep}".join(compare_to)

    write_to = [CFD, "test_files", f"{pdb_id}_test.pdb"]
    write_to = f"{os.sep}".join(write_to)

    write_pdb(pdb, file_name=write_to)
    with open(compare_to, "r", encoding="utf-8") as cf, open(
        write_to,
        "r",
        encoding="utf-8",
    ) as wf:
        next(cf)
        next(wf)
        cf_lines = cf.readlines()
        wf_lines = wf.readlines()
        assert cf_lines == wf_lines
    os.remove(write_to)



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""MOL2 format writing.

Write a dict of ``Pandas DataFrame`` back to a MOL2 file.

Currently, only the ``MOLECULE``, ``ATOM``, and ``BOND`` categories can be written back.

"""
from __future__ import annotations

import os
from datetime import date

import pandas as pd  # type: ignore

from .version import __version__ as pdbx2df_version

IMPLEMENTED_MOL2_CATS = ["MOLECULE", "ATOM", "BOND", "HEADER"]
"""MOL2 categories that are currently implemented."""


def write_mol2(
    mol2: dict[str, pd.DataFrame],
    file_name: str | os.PathLike | None = None,
) -> None:
    """Write a dict of ``Pandas DataFrame`` s into a MOL2 file.
    See https://is.muni.cz/th/fzk5s/dp_jakub_Vana.pdf p19 for column definitions.

    Args:
        mol2 (required): a ``dict`` of ``Pandas DataFrame`` s to write.
        file_name (optional): file name to write a MOL2 file. If ``None``,
            ``pdbx2df_output.mol2`` will be used as the file name.
            Defaults to **None**.

    Raises:
        TypeError: if ``mol2`` is not a valid dict of ``DataFrame``.
        ValueError: if the ``mol2`` contains other than supported categories.
    """
    if not file_name:
        file_name = "pdbx2df_output.mol2"

    if not isinstance(mol2, dict):
        raise TypeError(f"'mol2' has to be a dict but {type(mol2)} is provided.")

    implemented = ", ".join(IMPLEMENTED_MOL2_CATS)
    for key in mol2:
        if key not in IMPLEMENTED_MOL2_CATS:
            raise ValueError(f"Only {implemented} are implemented for the MOL2 format.")

        if not isinstance(mol2[key], pd.DataFrame):
            raise TypeError(f"{mol2[key]} is not a Pandas DataFrame object.")

    with open(file_name, "w", encoding="utf-8") as out_file:
        out_file.write("###\n")
        today = date.today().strftime("%Y-%m-%d")
        out_file.write(f"### Created by pdbx2df v{pdbx2df_version} {today}\n")

        if "HEADER" in mol2:
            df_header = mol2["HEADER"]
            for col_name in df_header.columns:
                header_line = df_header[col_name].to_list()[0]
                if col_name.startswith("info_"):
                    out_file.write(f"### Original header: {header_line}\n")
                else:
                    out_file.write(f"### {col_name}: {header_line}\n")

        out_file.write("###\n\n")
        out_file.write("@<TRIPOS>MOLECULE\n")
        if "MOLECULE" in mol2:
            df_molecule = mol2["MOLECULE"]
            mol_name = df_molecule.mol_name.to_list()[0]
            num_atoms = df_molecule.num_atoms.to_list()[0]
            num_bonds = df_molecule.num_bonds.to_list()[0]
            num_subst = df_molecule.num_subst.to_list()[0]
            num_feat = df_molecule.num_feat.to_list()[0]
            num_sets = df_molecule.num_sets.to_list()[0]
            mol_type = df_molecule.mol_type.to_list()[0]
            charge_type = df_molecule.charge_type.to_list()[0]

            status_bits, mol_comment = "", ""
            if status_bits in df_molecule.columns:
                status_bits = df_molecule.status_bits.to_list()[0]
                if mol_comment in df_molecule.columns:
                    mol_comment = df_molecule.mol_comment.to_list()[0]
            out_file.write(f"{mol_name}\n")
            out_file.write(
                f" {num_atoms} {num_bonds} {num_subst} {num_feat} {num_sets}\n"
            )
            out_file.write(f"{mol_type}\n")
            out_file.write(f"{charge_type}\n")
            if status_bits:
                out_file.write(f"{status_bits}\n")
                if mol_comment:
                    out_file.write(f"{mol_comment}\n")

        if "ATOM" in mol2:
            out_file.write("\n@<TRIPOS>ATOM\n")
            for _, row in mol2["ATOM"].iterrows():
                atom_id = row["atom_id"]
                atom_name = row["atom_name"]
                x = row["x"]
                y = row["y"]
                z = row["z"]
                atom_type = row["atom_type"]
                try:
                    subst_id = row["subst_id"]
                except KeyError:
                    subst_id = ""
                try:
                    subst_name = row["subst_name"]
                except KeyError:
                    subst_name = ""
                try:
                    charge = row["charge"]
                except KeyError:
                    charge = ""
                try:
                    status_bit = row["status_bit"]
                except KeyError:
                    status_bit = ""

                line = f"{atom_id:>6d} {atom_name:<8s} {x:>10.4f} {y:>10.4f} {z:>10.4f}"
                line += f" {atom_type:<9s} {subst_id:<2d} {subst_name:<7s} "
                line += f"{charge:>10.4f} {status_bit}"
                line = line.rstrip()
                out_file.write(line + "\n")

        if "BOND" in mol2:
            out_file.write("@<TRIPOS>BOND\n")
            for _, row in mol2["BOND"].iterrows():
                bond_id = row["bond_id"]
                origin_atom_id = row["origin_atom_id"]
                target_atom_id = row["target_atom_id"]
                bond_type = row["bond_type"]
                try:
                    status_bit = row["status_bit"]
                except KeyError:
                    status_bit = ""
                line = f"{bond_id:>6d} {origin_atom_id:>6d} {target_atom_id:>6d} "
                line += f"    {bond_type:1s} {status_bit:s}"
                line = line.rstrip()
                out_file.write(line + "\n")



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""PDBx/mmCIF format reading.

Reads a PDBx file into a dict of ``Pandas DataFrame`` s.
The dict keys are read from the mmCIF category names automatically.
For each category, the column names of the corresponding ``DataFrame``
are read form the category attributes automatically.

For example:
::

  _audit_conform.dict_name       mmcif_pdbx.dic
  _audit_conform.dict_version    5.355
  _audit_conform.dict_location   http://mmcif.pdb.org/dictionaries/ascii/mmcif_pdbx.dic

In this category, the category name is ``_audit_conform``, so the returned dict
key is ``_audit_conform``. The category attributes are ``dict_name``,
``dict_version``, and ``dict_location``, so the returned dict value as a
``DataFrame`` has the exact column names.
"""
from __future__ import annotations

import io
import os
import urllib.request
from collections import defaultdict
from pathlib import Path

import pandas as pd  # type: ignore

from .split_line import split_line

AF2_MODEL = 4
"""For AlphaFold structures, the version to use."""


def read_pdbx(
    pdbx_file: str | os.PathLike[str] | None = None,
    pdb_id: str | None = None,
    save_pdbx_file: bool = True,
    pdbx_file_dir: str | os.PathLike | None = None,
    category_names: list | None = None,
) -> dict[str, pd.DataFrame]:
    """Reads a ``.cif`` file's categories into a ``dict`` of ``Pandas DataFrame`` s.

    Args:
        pdb_id (optional): PDB/Uniprot ID. Required if ``pdbx_file`` is ``None``.
            Defaults to **None**.
        pdbx_file (optional): file name for a PDBx/mmCIF file. Used over `pdb_id`.
            Defaults to **None**.
        category_names (optional): a list of categories in the mmCIF file format.
            If ``None``, ``all`` is used and all categories will be processed.
            Defaults to **None**.
        save_pdbx_file (optional): whether to save the fetched PDBx file from RCSB
            to ``pdbx_file_dir``. Defaults to **False**.
        pdbx_file_dir (optional): directory to save fetched PDBx files. If ``None`` but
            ``save_pdbx_file`` is ``True``, the current working directory is used.
            Defaults to **None**.

    Returns:
        A dict of ``Pandas DataFrame`` s corresponding to required categories.

    Raises:
        ValueError: if none of ``pdb_id`` or ``pdbx_file`` is provided, or if ``pdb_id``
            is given but cannot the PDB file cannot be downloaded from RCSB, or the
            PDB file is corrupted like no end-line symbol, or some content is irregular.
        FileNotFoundError: if ``pdbx_file`` cannot be found.
    """
    data: dict[str, pd.DataFrame] = {}
    if pdb_id is None and pdbx_file is None:
        raise ValueError("At least one of pdb_id and pdbx_file has to be given.")
    elif pdbx_file is None:
        pdb_id = str(pdb_id).upper()
        if len(pdb_id) == 4:
            pdbx_file_url = f"https://files.rcsb.org/view/{pdb_id.upper()}.cif"
        else:
            pdbx_file_url = f"https://alphafold.ebi.ac.uk/files/AF-{pdb_id.upper()}-F1-model_v{AF2_MODEL}.cif"
        try:
            with urllib.request.urlopen(pdbx_file_url) as response:
                raw_data = response.read()
            text = raw_data.decode("utf-8")
            pdbx_file_handle: io.TextIOWrapper | io.StringIO = io.StringIO(text)
            if save_pdbx_file:
                if pdbx_file_dir is None:
                    pdbx_file_dir = "./PDBx_files"
                pdbx_file_dir = Path(pdbx_file_dir)
                if not pdbx_file_dir.exists():
                    pdbx_file_dir.mkdir(parents=True, exist_ok=True)
                file_path = Path(pdbx_file_dir, f"{pdb_id}.cif")
                with open(file_path, "w", encoding="utf-8") as p_file:
                    p_file.write(text)
        except urllib.error.HTTPError as http_error:
            raise ValueError(
                f"Cannot download PDBx file from url {pdbx_file_url}."
            ) from http_error
    else:
        pdbx_file = Path(pdbx_file)
        if not pdbx_file.exists():
            raise FileNotFoundError(f"File {pdbx_file} not found.")
        pdbx_file_handle = open(pdbx_file, "r", encoding="utf-8")
    if not category_names:
        category_names = ["all"]
    category_name = ""
    with pdbx_file_handle:
        line = pdbx_file_handle.readline()  # title line
        _ = line.strip().split("_")[1]
        processing_category = False
        while line:
            if line[0] == "#":
                processing_category = True
                loop_category = False
                category_lines = ""
                category_cols = []
                category_dict = defaultdict(list)
                line = pdbx_file_handle.readline()  # category first line
                if not line:
                    break
                if line == "loop_\n":
                    loop_category = True
                    line = pdbx_file_handle.readline()
                in_multiple_line = (
                    False  # lines quoted by ';', they need to be specially treated
                )
                while line.strip() not in ["#", "#   #", "##"]:
                    if line[0] == "_" and not in_multiple_line:
                        category_name, category_nth_col = (
                            line.strip().split(" ")[0].split(".")
                        )
                        category_cols.append(category_nth_col)
                    if (
                        category_name in category_names or "all" in category_names
                    ):  # in a required category
                        if line[0] == ";":
                            in_multiple_line = True  # a ';' quoted record begins
                            line = line.replace(
                                '"', "'"
                            )  # to avoid conflict with the next action
                            line = '"' + line[1:]
                            # these ; symbols are not contents but a pair of container
                            # symbols for a single record
                            # but other ';' are parts of a single record, so we use "
                            # to make it easier to parse
                            if len(line.strip()) == 1:
                                in_multiple_line = False  # a ';' quoted record ends
                        elif in_multiple_line:
                            line = line.replace(
                                '"', "'"
                            )  # to avoid conflict with the above action
                        category_lines += line
                    line = pdbx_file_handle.readline()
                    if not line:
                        raise ValueError(
                            f"{pdbx_file} not normally ended. Download it again?"
                        )
                if "all" not in category_names and category_name not in category_names:
                    continue
                category_lines = category_lines.replace(
                    "\n", " "
                )  # cannot rely on newline to parse anyway.
                records = split_line(category_lines)
                if not loop_category:
                    if len(records) != 2 * len(category_cols):
                        raise ValueError(
                            f"{pdbx_file} category {category_name} has irregular contents: {records}"  # noqa
                        )
                    for i, rec in enumerate(records):
                        if i % 2 == 0:
                            col_name = rec.split(".")[1]
                            if col_name in [
                                "pdbx_seq_one_letter_code",
                                "pdbx_seq_one_letter_code_can",
                            ]:
                                records[i + 1] = records[i + 1].replace(" ", "")
                            category_dict[rec.split(".")[1]] = [records[i + 1]]
                else:
                    if (
                        records[len(category_cols)][0] == "_"
                        or records[len(category_cols) - 1][0] != "_"
                        or not len(records) % len(category_cols) == 0
                    ):
                        raise ValueError(
                            f"""{pdbx_file} category {category_name} has irregular contents:
                            {len(records)} for {len(category_cols)}"""  # noqa
                        )
                    record_data = records[len(category_cols) :]
                    for i, i_data in enumerate(record_data):
                        col = i % len(category_cols)
                        if category_cols[col] in [
                            "pdbx_seq_one_letter_code",
                            "pdbx_seq_one_letter_code_can",
                        ]:
                            i_data = i_data.replace(
                                " ", ""
                            )  # delete the gaps in the one_letter_code
                        category_dict[category_cols[col]].append(i_data)
                data[category_name] = pd.DataFrame(category_dict)
                for c_name in category_names:
                    if c_name not in data:
                        break
                        # at least one required_category is not processed,
                        # continue the main while loop
                else:
                    break
                    # all required_categories are processed,
                    # break the main while loop

            if not processing_category:
                line = pdbx_file_handle.readline()
    if "_atom_site" in data:
        col_dtypes = {
            "id": "int",
            "label_entity_id": "int",
            "label_seq_id": "int",
            "Cartn_x": "float",
            "Cartn_y": "float",
            "Cartn_z": "float",
            "occupancy": "float",
            "B_iso_or_equiv": "float",
            "auth_seq_id": "int",
            "pdbx_PDB_model_num": "int",
            "pdbx_sifts_xref_db_num": "int",
        }
        col_dtypes = {
            col_name: data_type
            for col_name, data_type in col_dtypes.items()
            if col_name in data["_atom_site"]
        }

        data["_atom_site"] = data["_atom_site"].astype(col_dtypes)
    return data



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""PDB format writing.

Write a dict of ``Pandas DataFrame`` back to a PDB file.

Currently, only the ``_atom_site`` category can be written back.

"""
from __future__ import annotations

import os
import warnings
from datetime import date

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from .version import __version__ as pdbx2df_version

IMPLEMENTED_PDB_CATS = ["_atom_site"]
"""PDB categories that are currently implemented."""


def write_pdb(
    pdb: dict[str, pd.DataFrame],
    file_name: str | os.PathLike | None = None,
    allow_chimera: bool = False,
) -> None:
    """Write a dict of ``Pandas DataFrame`` s into a PDB file.

    Args:
        pdb (required): a ``dict`` of ``Pandas DataFrame`` s to write.
        file_name (optional): file name to write a PDB file. If ``None``,
            ``pdbx2df_output.pdb`` will be used as the file name.
            Defaults to **None**.
        allow_chimera (optional): whether to allow writing to Chimera-formatted PDB
            files. Defaults to **False**.

    Raises:
        TypeError: if ``pdb`` is not a valid dict of ``DataFrame``.
        ValueError: if the ``pdb`` contains other than supported categories.
    """
    if not file_name:
        file_name = "pdbx2df_output.pdb"

    if not isinstance(pdb, dict):
        raise TypeError(f"pdb has to be a dict but {type(pdb)} is provided.")

    implemented = ", ".join(IMPLEMENTED_PDB_CATS)
    for key in pdb.keys():
        if key not in IMPLEMENTED_PDB_CATS:
            raise ValueError(f"Only {implemented} are implemented for the PDB format.")

    with open(file_name, "w", encoding="utf-8") as out_file:
        today = date.today().strftime("%Y-%m-%d")
        padding = " "
        tag = f"CREATED WITH PDBX2DF v{pdbx2df_version} {today}  "
        header = f"REMARK   1{tag:>70}\n"
        out_file.write(header)

        if "_atom_site" in pdb.keys():
            df_atom_site = pdb["_atom_site"]
            str_names = [
                "atom_name",
                "alt_loc",
                "residue_name",
                "chain_id",
                "insertion",
                "segment_id",
                "element_symbol",
                "charge",
            ]
            df_atom_site[str_names] = df_atom_site[str_names].fillna("")
            n_nmr_models = 0
            if "nmr_model" in df_atom_site.columns:
                n_nmr_models = len(df_atom_site.nmr_model.unique())
                out_file.write(f"NUMMDL{n_nmr_models:>6}{padding:>68}\n")
            prev_nmr_model = None
            n_nmr_written = 0
            for _, row in df_atom_site.iterrows():
                record_name = row[
                    "record_name"
                ]  # 1-6 (normal) or 1-5 (Chimera);char;left
                atom_number = row[
                    "atom_number"
                ]  # 7-11 (normal) or 6-11 (Chimera);char;left
                # 12 (blank)
                atom_name = row[
                    "atom_name"
                ]  # 13-16 (2/1-l elements left-13/14 unless == 4);char
                alt_loc = row["alt_loc"]  # 17;char
                residue_name = row[
                    "residue_name"
                ]  # 18-20 (normal) or 18-21 (Chimera);char;right
                # 21 (blank) or NA (Chimera)
                chain_id = row["chain_id"]  # 22;char
                residue_number = row["residue_number"]  # 23-26;int;right
                insertion = row["insertion"]  # 27;char
                # 28-30 (blank)
                x_coord = row["x_coord"]  # 31-38 (8.3);right
                y_coord = float(row["y_coord"])  # 39-46 (8.3);right
                z_coord = float(row["z_coord"])  # 47-54 (8.3);right
                occupancy = float(row["occupancy"])  # 55-60 (6.2);right
                b_factor = float(row["b_factor"])  # 61-66 (6.2);right
                # 67-72 (blank)
                segment_id = row["segment_id"]  # 73-76;char;left
                element_symbol = row["element_symbol"]  # 77-78;char;right
                charge = row["charge"]  # 79-80;char(?);right?
                if n_nmr_models > 0:
                    nmr_model = row["nmr_model"]
                    if prev_nmr_model is None:
                        out_file.write(f"MODEL {nmr_model:>6}{padding:>68}\n")
                    elif nmr_model != prev_nmr_model:
                        out_file.write(f"ENDMDL{padding:>74}\n")
                        n_nmr_written += 1
                        out_file.write(f"MODEL {nmr_model:>6}{padding:>68}\n")
                    prev_nmr_model = nmr_model
                if len(element_symbol) == 1 and len(atom_name) < 4:
                    atom_name = " " + atom_name
                if allow_chimera:
                    if len(record_name) == 6:
                        warnings.warn(
                            f"Record name {record_name} length was 6 and is truncated to {record_name[:-1]}",  # noqa
                            RuntimeWarning,
                            stacklevel=2,
                        )
                        record_name = record_name[:-1]
                    atom_site_line = f"{record_name:<5s}{atom_number:>6d} {atom_name:<4s}{alt_loc:<1s}"  # noqa
                    if len(residue_name) < 4:
                        residue_name = residue_name + " "
                    atom_site_line += f"{residue_name:>4s}{chain_id:<1s}{residue_number:>4d}{insertion:<1s}   "  # noqa
                else:
                    atom_site_line = f"{record_name:<6s}{atom_number:>5d} {atom_name:<4s}{alt_loc:<1s}"  # noqa
                    atom_site_line += f"{residue_name:>3s} {chain_id:<1s}{residue_number:>4d}{insertion:<1s}   "  # noqa
                if np.isnan(x_coord):
                    x_coord_e, y_coord_e, z_coord_e, occupancy_e, b_factor_e = (
                        " ",
                        " ",
                        " ",
                        " ",
                        " ",
                    )
                    atom_site_line += f"{x_coord_e:>8s}{y_coord_e:>8s}{z_coord_e:>8s}{occupancy_e:>6s}"  # noqa
                    atom_site_line += f"{b_factor_e:>6s}      {segment_id:<4s}{element_symbol:>2s}{charge:>2s}\n"  # noqa
                else:
                    atom_site_line += (
                        f"{x_coord:8.3f}{y_coord:8.3f}{z_coord:8.3f}{occupancy:6.2f}"
                    )
                    atom_site_line += f"{b_factor:6.2f}      {segment_id:<4s}{element_symbol:>2s}{charge:>2s}\n"  # noqa
                out_file.write(atom_site_line)
            if n_nmr_models > 0:
                out_file.write(f"ENDMDL{padding:>74}\n")



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Test for reading PDBx/mmCIF files."""
import os
import sys

import pandas as pd  # type: ignore

from pdbx2df.read_pdbx import read_pdbx

sys.path.append("..")
CFD = os.path.dirname(__file__)


def test_read_pdbx():
    """
    Test read_pdbx function
    """
    file_path = [CFD, "test_files", "1VII.cif"]
    test_file = f"{os.sep}".join(file_path)

    # Basic one category read
    pdbx = read_pdbx(pdbx_file=test_file, category_names=["_entry"])
    expected = {"_entry": pd.DataFrame({"id": ["1VII"]})}
    pd.testing.assert_frame_equal(pdbx["_entry"], expected["_entry"])

    # Multiple categories read
    pdbx = read_pdbx(pdbx_file=test_file, category_names=["_entry", "_entity_name_com"])
    expected = {
        "_entry": pd.DataFrame({"id": ["1VII"]}),
        "_entity_name_com": pd.DataFrame(
            {"entity_id": ["1"], "name": ["HP-36, R42-76"]}
        ),
    }
    pd.testing.assert_frame_equal(pdbx["_entry"], expected["_entry"])
    pd.testing.assert_frame_equal(
        pdbx["_entity_name_com"], expected["_entity_name_com"]
    )

    # category with ';'
    pdbx = read_pdbx(pdbx_file=test_file, category_names=["_pdbx_nmr_refine"])
    expected = {
        "_pdbx_nmr_refine": pd.DataFrame(
            {
                "entry_id": ["1VII"],
                "method": ["DISTANCE GEOMETRY SIMULATED ANNEALING"],
                "details": [
                    "THE X-PLOR (R6)1/6 NOE POTENTIAL WAS USED FOR NOE'S INVOLVING NON-STEREOSPECIFICALLY ASSIGNED METHYL, METHYLENE, AND AROMATIC PROTONS. NO ATTRACTIVE POTENTIALS WERE USED IN CALCULATING THE STRUCTURES. THE VAN DER WAALS CUTOFF USED FOR THE X-PLOR REPEL FUNCTION WAS 0.75 ANGSTROMS. AFTER DISTANCE GEOMETRY AND REGULARIZATION, EACH STRUCTURE WAS SUBJECTED TO ONE ROUND OF SIMULATED ANNEALING FROM 2000K TO 100K OVER 2000 STEPS. THIS IS THE AVERAGE OF 29 STRUCTURES MINIMIZED USING ONLY REPULSIVE POTENTIALS. "  # noqa
                ],
                "software_ordinal": ["1"],
            }
        )
    }
    pd.testing.assert_frame_equal(
        pdbx["_pdbx_nmr_refine"], expected["_pdbx_nmr_refine"]
    )

    # From PDB ID
    pdb_id = "1VIi"
    pdbx = read_pdbx(
        pdb_id=pdb_id,
        category_names=["_atom_site"],
        save_pdbx_file=True,
    )
    pdb_df = pdbx["_atom_site"]
    assert os.path.exists(
        f"./PDBx_files/{pdb_id.upper()}.cif"
    ), "File not saved to PDBx_files if pdbx_file_dir not provided."
    assert (
        pdb_df.head(1).label_comp_id == "MET"
    ).item(), "First 1VII residue is not MET."

    # From Uniprot ID
    uniprot_id = "P01116"
    pdbx = read_pdbx(
        pdb_id=uniprot_id,
        category_names=["_atom_site"],
        save_pdbx_file=True,
    )
    pdb_df = pdbx["_atom_site"]
    assert os.path.exists(
        f"./PDBx_files/{uniprot_id.upper()}.cif"
    ), "File not saved to PDBx_files if pdbx_file_dir not provided."
    assert (
        pdb_df.head(1).label_seq_id == 1
    ).item(), "First P01116 residue number is not 1."



import os

__package_name__ = "pdbx2df"
__author__ = "Ruibin Liu"

from setuptools import find_packages, setup  # type: ignore

if os.path.exists("README.md"):
    long_description = open("README.md").read()
else:
    long_description = "pdbx2df - Super lightweight and fast mmCIF/PDB/MOL2 file parser into Pandas DataFrames and backwards writer."  # noqa

with open("requirements.txt") as f:
    REQUIREMENTS = f.read().strip().split("\n")

ver = {}  # type: ignore
with open("pdbx2df/version.py", "r") as vf:
    exec(vf.read(), ver)

setup(
    name=__package_name__,
    version=ver["__version__"],
    author=__author__,
    author_email="ruibinliuphd@gmail.com",
    description="Super lightweight and fast mmCIF/PDB/MOL2 file parser into Pandas DataFrames and backwards writer.",  # noqa
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ruibin-Liu/pdbx2df",
    project_urls={
        "Bug Tracker": "https://github.com/Ruibin-Liu/pdbx2df/issues",
        "Documentation": "https://pdbx2df.readthedocs.io/en/stable/",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    python_requires=">=3.7",
)



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Tests for writing PDBx/mmCIF files."""
import filecmp
import os
import sys

from pdbx2df.read_pdbx import read_pdbx
from pdbx2df.write_pdbx import write_pdbx

sys.path.append("..")
CFD = os.path.dirname(__file__)


def test_write_pdbx():
    """
    Test write_pdbx function
    """
    test_file = [CFD, "test_files", "1VII.cif"]
    test_file = f"{os.sep}".join(test_file)
    pdbx = read_pdbx(pdbx_file=test_file, category_names=["all"])

    compare_to = [CFD, "test_files", "1VII_pdbx2df.cif"]
    compare_to = f"{os.sep}".join(compare_to)

    write_to = [CFD, "test_files", "1VII_test.cif"]
    write_to = f"{os.sep}".join(write_to)

    write_pdbx(pdbx, file_name=write_to)
    assert filecmp.cmp(compare_to, write_to)
    os.remove(write_to)



# pdbx2df
# Author: Ruibin Liu <ruibinliuphd@gmail.com>
# License: MIT
# Code Repository: https://github.com/Ruibin-Liu/pdbx2df
"""Tests for reading mol2 files."""
import os
import sys

import pytest

from pdbx2df.read_mol2 import read_mol2

sys.path.append("..")
CFD = os.path.dirname(__file__)


def test_read_mol2():
    """
    Test read_mol2 function
    """
    # Correct ones
    file_path = [CFD, "test_files", "test.mol2"]
    test_file = f"{os.sep}".join(file_path)
    mol2 = read_mol2(test_file)
    mol2_atom_df = mol2["ATOM"]
    mol2_bond_df = mol2["BOND"]
    mol2_molecule_df = mol2["MOLECULE"]
    assert (mol2_atom_df.head(1).atom_id == 1).item(), "The first atom_id is not 1."
    assert (mol2_bond_df.tail(1).bond_id == 62).item(), "The last bond_id is not 62."
    assert (
        mol2_molecule_df.tail(1).mol_name == "KEGG_DURG-00000915-01"
    ).item(), "The MOLECULE mol_name is not KEGG_DURG-00000915-01."

    # Not implemented
    message = """Only ATOM, MOLECULE, BOND, HEADER categories are implemented for the MOL2 format.
                Create an issue at https://github.com/Ruibin-Liu/pdbx2df if
                you want the ANCHOR_ATOM category to be implemented.
                """  # noqa
    with pytest.raises(NotImplementedError) as error_msg:
        _ = read_mol2(test_file, category_names=["ANCHOR_ATOM"])
    assert message in str(error_msg.value), "NotImplementedError not raised correctly."
